import numpy as np
import pandas as pd
from zipfile import ZipFile
from scipy.io import loadmat
from mne.channels import make_standard_montage
from mne import create_info
from mne.io import RawArray

from .base import RawDataset, DATA_PATH
from .download import data_dl
from ..utils import get_init_args
from ..tools.docs import fill_doc


URL = "https://reshare.ukdataservice.ac.uk/"


@fill_doc
class MODMA_128_Resting(RawDataset):
    """Multi-modal Open Dataset for Mental-disorder Analysis, Experimental Data.

    .. admonition:: Dataset summary

        ====== ====== ======= ===== ======== ======
        Subj   Chan   Time    Cls   Freq     Sess
        ====== ====== ======= ===== ======== ======
        53     128    4.5 s   2     250 Hz   1
        ====== ====== ======= ===== ======== ======

    Multi-model open dataset for mental-disorder analysis [1]_. The dataset
    includes data mainly from clinically depressed patients and matching normal
    controls. 53 participants include a total of 24 outpatients (13 males and
    11 females; 16-56-year-old) diagnosed with depression, as well as 29 healthy
    controls (20 males and 9 females; 18-55-year-old) were recruited. No
    experimental material. The participants should keep quiet and close their
    eyes as much as possible. Continuous EEG signals were recorded using a
    128-channel HydroCel Geodesic Sensor Net (Electrical Geodesics Inc., Oregon
    Eugene, USA) and Net Station acquisition software (version 4.5.4). The
    sampling frequency was 250 Hz. All raw electrode signals were referenced
    to the Cz. 5 minutes of eyes-closed resting-state EEG was recorded.
    Participants were required to keep awake and still without any bodily
    movements, including heads or legs, and any unnecessary eye movements,
    saccades, and blinks.

    References
    ----------
    .. [1] A. Seal, R. Bajpai, J. Agnihotri, A. Yazidi, E. Herrera-Viedma, and
        O. Krejcar, DeprNet: A deep convolution neural network framework for
        detecting depression using EEG, IEEE Trans. Instrum. Meas., vol. 70,
        pp. 1-13, 2021, doi: 10.1109/TIM.2021.3053999.

    Parameters
    ----------
    %(subjects)s
    %(raw_tmin_tmax)s
    %(picks)s
    %(resample)s
    %(rename)s
    """

    def __init__(
        self,
        subjects: list[int] | None = None,
        tmin: float = 0,
        tmax: float | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        rename: str | None = None,
    ) -> None:
        if picks is None:
            picks = [f"E{i}" for i in range(1, 129)]

        super().__init__(
            repr=get_init_args(self, locals(), rename=rename, ret_dict=True),
            subject_list=list(range(1, 54)),
            event_id={"major_depressive_disorder": 1, "healthy_controls": 2},
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            picks=picks,
            resample=resample,
        )
        self._data_url = (
            f"{URL}854301/4/854301_EEG_128Channels_Resting_Lanzhou_2015.zip"
        )
        self._data_path = DATA_PATH / "modma"
        self._montage = make_standard_montage("GSN-HydroCel-129")
        self._info = create_info(
            ch_names=self._montage.ch_names, sfreq=250.0, ch_types="eeg"
        )

    def _parse_zip(self):
        path_zip = data_dl(self._data_url, self._data_path)

        # Extract the zip file if it hasn't been extracted yet
        path_folder = self._data_path / "EEG_128channels_resting_lanzhou_2015"
        if not path_folder.exists():
            print("The first read requires decompressing all data.")
            with ZipFile(path_zip, "r") as zipf:
                zipf.extractall(self._data_path)

        return path_folder

    def encoding(self):
        """Return the correspondence between subjects and source files within
        the datasets.
        """
        path_folder = self._parse_zip()
        df = pd.read_excel(
            path_folder
            / "subjects_information_EEG_128channels_resting_lanzhou_2015.xlsx",
            usecols="A:K",
        )
        df.index += 1
        return df

    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        path_folder = self._parse_zip()
        path_list = sorted(path_folder.glob("*.mat"))

        path_subject = path_list[subject - 1]
        data = loadmat(path_subject, struct_as_record=False, squeeze_me=True)
        eeg_data = data[list(data.keys())[3]]
        eeg_data *= 1e-9
        raw = RawArray(data=eeg_data, info=self._info, verbose=verbose)
        raw.set_montage(self._montage)

        return {"session_1": {"run_1": raw}}

    def _set_label(self, subject: int):
        encoding = self.encoding()
        subject_type = encoding.loc[subject, "type"]

        if subject_type == "MDD":
            return np.array([1])
        else:
            return np.array([2])



# The code for downloading datasets is adapted from the MOABB project (licensed
# under the BSD 3-Clause License). Modifications were made to align with the
# data processing methods used in the dpeeg project.

from .base import (
    EEGData,
    MultiSessEEGData,
    SplitEEGData,
    EEGDataset,
)
from .loaddataset import LoadDataset
from .utils import save, load, save_dataset
from .bnci import BCICIV2A, BCICIV2B
from .gigadb import OpenBMI_MI
from .gin import HighGamma
from .physionet import PhysioNet_MI
from .ukds import MODMA_128_Resting



import torch
import torch.nn as nn

from .utils import Conv2dWithNorm, LinearWithNorm
from ..tools.docs import fill_doc


__all__ = ["EEGNet"]


@fill_doc
class EEGNet(nn.Module):
    """EEGNet: A Compact Convolutional Neural Network for EEG-based
    Brain-Computer Interfaces (EEGNet).

    EEGNet [1]_ is a compact convolutional neural network for EEG-based BCIs.
    EEGNet starts with a temporal convolution to learn frequency filters, then
    uses a depthwise convolution, connected to each feature map individually,
    to learn frequency-specific spatial filters. The separable convolution is a
    combination of a depthwise convolution, which learns a temporal summary for
    each feature map individually, followed by a pointwise convolution, which
    learns how to optimally mix the feature maps together.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    F1 : int
        Number of temporal filters.
    C1 : int
        Temporal convolution kernel size.
    D : int
        Depth of depthwise convolution.
    F2 : int
        Number of separable convolutions.
    C2 : int
        Separable convolution kernel size.
    P1 : int
        The first pooling kernel size.
    P2 : int
        The second pooling kernel size.
    dropout : float
        Dropout rate.

    References
    ----------
    .. [1] V. J. Lawhern, A. J. Solon, N. R. Waytowich, S. M. Gordon,
        C. P. Hung, and B. J. Lance, “EEGNet: a compact convolutional neural
        network for EEG-based brain–computer interfaces,” J. Neural Eng.,
        vol. 15, no. 5, p. 056013, Jul. 2018, doi: 10.1088/1741-2552/aace8c.
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        F1: int = 8,
        C1: int = 63,
        D: int = 2,
        F2: int = 16,
        C2: int = 15,
        P1: int = 8,
        P2: int = 16,
        dropout: float = 0.5,
    ) -> None:
        super().__init__()

        self.filter = nn.Sequential(
            nn.Conv2d(1, F1, (1, C1), padding=(0, C1 // 2), bias=False),
            nn.BatchNorm2d(F1),
        )

        self.depthwise_conv = nn.Sequential(
            Conv2dWithNorm(F1, D * F1, (nCh, 1), groups=F1, bias=False, max_norm=1),
            nn.BatchNorm2d(D * F1),
            nn.ELU(),
            nn.AvgPool2d((1, P1)),
            nn.Dropout(dropout),
        )

        self.separable_conv = nn.Sequential(
            nn.Conv2d(
                D * F1, D * F1, (1, C2), padding=(0, C2 // 2), groups=D * F1, bias=False
            ),
            nn.Conv2d(D * F1, F2, 1, bias=False),
            nn.BatchNorm2d(F2),
            nn.ELU(),
            nn.AvgPool2d((1, P2)),
            nn.Dropout(dropout),
        )

        self.flatten = nn.Flatten()
        self.fc = nn.Sequential(
            # Experimental results show that using linearwithnorm will lead to
            # performance degradation.
            # LinearWithNorm(self.get_size(nCh, nTime), nCls, bias=True, max_norm=0.25)
            nn.Linear(self.get_size(nCh, nTime), nCls, bias=True),
            nn.LogSoftmax(dim=1),
        )

    def get_size(self, nCh, nTime):
        x = torch.randn(1, 1, nCh, nTime)
        out = self.filter(x)
        out = self.depthwise_conv(out)
        out = self.separable_conv(out)
        return self.flatten(out).size(1)

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, 1, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        """
        out = self.filter(x)
        out = self.depthwise_conv(out)
        out = self.separable_conv(out)
        out = self.flatten(out)
        return self.fc(out)



import torch
import torch.nn as nn
from torch.optim.adamw import AdamW
from torch.nn.init import trunc_normal_, constant_

from ..tools.docs import fill_doc


__all__ = ["IFNet", "IFNetAdamW"]


class InterFre(nn.Module):
    def forward(self, x):
        out = sum(x)
        out = nn.GELU()(out)
        return out


@fill_doc
class IFNet(nn.Module):
    """IFNet: An Interactive Frequency Convolutional Neural Network for
    Enhancing Motor Imagery Decoding From EEG (IFNet).

    Inspired by the concept of cross-frequency coupling and its correlation
    with different behavioral tasks, IFNet [1]_ explores cross-frequency
    interactions for enhancing representation of MI characteristics. IFNet
    first extracts spectro-spatial features in low and high-frequency bands,
    respectively. Then the interplay between the two bands is learned using an
    element-wise addition operation followed by temporal average pooling.
    Combined with repeated trial augmentation as a regularizer, IFNet yields
    spectro-spatiotemporally robust features for the final MI classification.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    F : int
        Number of spectro-spatial filters.
    C : int
        Spectro-spatial filter kernel size.
    radix : int
        Number of cross-frequency domains.
    P : int
        Pooling kernel size.
    dropout : float
        Dropout rate.

    References
    ----------
    .. [1] J. Wang, L. Yao and Y. Wang, "IFNet: An Interactive Frequency
        Convolutional Neural Network for Enhancing Motor Imagery Decoding from
        EEG," in IEEE Transactions on Neural Systems and Rehabilitation
        Engineering, doi: 10.1109/TNSRE.2023.3257319.
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        F: int = 64,
        C: int = 63,
        radix: int = 2,
        P: int = 125,
        dropout: float = 0.5,
    ) -> None:
        super().__init__()
        self.F = F
        self.mF = F * radix

        self.sConv = nn.Sequential(
            nn.Conv1d(nCh * radix, self.mF, 1, bias=False, groups=radix),
            nn.BatchNorm1d(self.mF),
        )

        self.tConv = nn.ModuleList()
        for _ in range(radix):
            self.tConv.append(
                nn.Sequential(
                    nn.Conv1d(F, F, C, 1, padding=C // 2, groups=F, bias=False),
                    nn.BatchNorm1d(F),
                )
            )
            C //= 2

        self.interFre = InterFre()
        self.downSamp = nn.Sequential(nn.AvgPool1d(P), nn.Dropout(dropout))
        self.fc = nn.Sequential(
            nn.Flatten(), nn.Linear(int(F * (nTime // P)), nCls), nn.LogSoftmax(dim=1)
        )

        self.apply(self.initParms)

    def initParms(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.01)
            if m.bias is not None:
                constant_(m.bias, 0)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm1d, nn.BatchNorm2d)):
            if m.weight is not None:
                constant_(m.weight, 1.0)
            if m.bias is not None:
                constant_(m.bias, 0)
        elif isinstance(m, (nn.Conv1d, nn.Conv2d)):
            trunc_normal_(m.weight, std=0.01)
            if m.bias is not None:
                constant_(m.bias, 0)

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, nCh * radix, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        """
        out = self.sConv(x)
        out = torch.split(out, self.F, dim=1)
        out = [m(x) for m, x in zip(self.tConv, out)]
        out = self.interFre(out)
        out = self.downSamp(out)
        return self.fc(out)


class IFNetAdamW(AdamW):
    """Customized AdamW Optimizer for IFNet.

    IFNetAdamW optimizer allows bias and weights based on certain parameters to
    not decay.

    Parameters
    ----------
    net : IFNet
        IFNet model instance.
    """

    def __init__(self, net: nn.Module, **kwargs) -> None:
        has_decay = []
        no_decay = []

        for name, param in net.named_parameters():
            if not param.requires_grad:
                continue
            if len(param.shape) == 1 or name.endswith(".bias"):
                no_decay.append(param)
            else:
                has_decay.append(param)
        params = [{"params": has_decay}, {"params": no_decay, "weight_decay": 0}]

        super().__init__(params, **kwargs)



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import os
import random
import inspect
from collections.abc import Iterable, Mapping
from typing import overload, Literal

import mne
import torch
import numpy as np


DPEEG_SEED: int = 42  # Life, the Universe and Everything
DPEEG_DIR = os.path.join(os.path.expanduser("~"), "dpeeg")


def set_log_level(verbose=None, return_old_level=False):
    """Set the logging level.

    Parameters
    ----------
    verbose : bool, str, int, or None
        The verbosity of messages to print. If a str, it can be either DEBUG,
        INFO, WARNING, ERROR, or CRITICAL. Note that these are for
        convenience and are equivalent to passing in logging.DEBUG, etc.
        For bool, True is the same as 'INFO', False is the same as 'WARNING'.
        If None, the environment variable MNE_LOGGING_LEVEL is read, and if
        it doesn't exist, defaults to INFO.
    return_old_level : bool
        If True, return the old verbosity level.

    Returns
    -------
    old_level : int
        The old level. Only returned if ``return_old_level`` is True.
    """
    return mne.utils.set_log_level(verbose, return_old_level)


def _set_random_seed(seed: int) -> None:
    """Set the seed for Python's built-in random module and numpy.

    Parameters
    ----------
    seed: int
        The random seed to use.
    """
    random.seed(seed)
    np.random.seed(seed)


def _set_torch_seed(seed: int) -> None:
    """Set the seed for Python's built-in random module and numpy.

    Parameters
    ----------
    seed: int
        The random seed to use.

    """
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


def set_seed(seed: int = DPEEG_SEED, ret_old_seed: bool = False) -> None | int:
    """Set global random seed for random, Numpy and PyTorch.

    Parameters
    ----------
    seed : int
        The random seed to use. If None, the environment variable `DPEEG_SEED`
        is read, defaults to 42.
    ret_old_seed : bool, optional
        If True, return the old seed.

    Returns
    -------
    old_seed : int, None
        The old seed. Only returned if `ret_old_seed` is True.
    """
    global DPEEG_SEED
    old_seed = DPEEG_SEED
    DPEEG_SEED = seed
    _set_random_seed(seed)
    _set_torch_seed(seed)
    if ret_old_seed:
        return old_seed


def unpacked(*args) -> list:
    """Positional arguments are unpacked as lists."""
    result = [X for X in args]
    return result


def mapping_to_str(mapping: Mapping, kv_symbol: str = "=", symbol: str = ", ") -> str:
    """Convert a mapping (e.g., dict) into a string format.

    Parameters
    ----------
    mapping : Mapping
        The mapping (e.g., dict) to be converted into a string.
        It should contain keys and values that can be converted to strings.
    kv_symbol : str
        The separator to be used between key and value in the resulting string.
    symbol : str
        The separator to be used between key-value pairs in the resulting
        string. Defaults to ", " (a comma followed by a space).
    """
    return symbol.join(f"{k}{kv_symbol}{v}" for k, v in mapping.items())


def iterable_to_str(iterable: Iterable, symbol: str = ", ") -> str:
    """Convert any iterable (e.g., list, tuple, set) into a string format.

    Parameters
    ----------
    iterable : Iterable
        The iterable to be converted into a string.
        It should contain elements that can be converted to strings.
    symbol : str
        The separator to be used between elements in the resulting string.
        Defaults to ", " (a comma followed by a space).
    """
    return symbol.join(map(str, iterable))


def align_text(prefix: str, text: str, only_head: bool = True) -> str:
    """Aligns text, adding a specified prefix to the beginning of line.

    Parameters
    ----------
    prefix : str
        The prefix to be added at the beginning of line.
    text : str
        The text to be aligned, including multiple lines.
    only_head : bool
        If false, prefix will be prepended to each line. Otherwise, spaces with
        a length of prefix will be added to other lines except the first line.

    Returns
    -------
    align_text : str
        The aligned text.

    Examples
    --------
    >>> print(align_text("+ ", text, only_head=False))
    + First line
    + Second line
    """
    return "\n".join(
        [
            (
                f"{prefix}{line}"
                if (i == 0) or (not only_head)
                else f"{' ' * len(prefix)}{line}"
            )
            for i, line in enumerate(text.splitlines())
        ]
    )


def _format_log_kv(key, value, spaces: int = 2) -> str:
    return f"[{key}:\n{align_text(' ' * spaces, str(value))}\n]"


def _format_log(input: dict, spaces: int = 2) -> str:
    formatted_items = "\n".join(
        align_text(f"{' ' * spaces}[{key}]: ", str(value))
        for key, value in input.items()
        if key != "_obj_name"
    )
    return f"[{input['_obj_name']}:\n{formatted_items}\n]"


def _format_rp(input: dict, kv_symbol: str = "=") -> str:
    obj_name = input.pop("_obj_name")
    formatted_args = mapping_to_str(input, kv_symbol=kv_symbol)
    input["_obj_name"] = obj_name
    return f"{obj_name}({formatted_args})"


@overload
def get_init_args(
    obj,
    locals: dict,
    format: str = "log",
    rename: str | None = None,
    ret_dict: Literal[False] = False,
    **kwargs,
) -> str:
    pass


@overload
def get_init_args(
    obj,
    locals: dict,
    format: str = "log",
    rename: str | None = None,
    ret_dict: Literal[True] = True,
    **kwargs,
) -> dict:
    pass


def get_init_args(
    obj,
    locals,
    format="log",
    rename=None,
    ret_dict=False,
    **kwargs,
) -> str | dict:
    """Get object initialization parameters.

    Assemble the parameters passed in by class initialization or function
    according to the specified format. The function can return either the
    assembled string or the unassembled dictionary for updating internal
    parameters.

    Parameters
    ----------
    obj : class, function
        The kind of class or function.
    locals : dict
        Local parameters in class.__init__ or function.
    format : str
        Parameters splicing format.
        - log: [name \\n\\t param1 \\n\\t param2 \\n\\t ...]
        - rp: name(param1, param2, ...)
    rename : str, optional
        User renamed name. If None, the default class or function name is used.
    ret_dict : bool
        If `True`, return the initialization parameters as a dictionary,
        ignoring the `format` parameter. By default, return a string in the
        specified format.
    **kwargs : dict, optional
        Additional parameters for format assembly.

    Returns
    -------
    str
        A string specifying the format.
    dict
        A dictionary containing initialization parameters.

    Examples
    --------
    When you want to get the parameters of a class and its methods. Note: It is
    best to use `self` as the `obj` parameter during initialization and avoid
    using the class to prevent class_name recognition errors of inherited classes.

    >>> class Base:
    ...     def __init__(self, param1, param2=2) -> None:
    ...         print(get_init_args(self, locals(), "rp"))
    ...     def method(self, param3):
    ...         print(get_init_args(self.method, locals(), "rp"))
    >>> class Sub(Base):
    ...     pass
    >>> obj_base = Base(1)
    Base(param1=1, param2=2)
    >>> obj_base.method(3)
    Base.method(param3=3)
    >>> obj_sub = Sub(1)
    Sub(param1=1, param2=2)
    >>> obj_sub.method(3)
    Sub.method(param3=3)

    Likewise, functions are supported.

    >>> def func(param1, param2 = 2):
    ...     print(get_init_args(func, locals(), "rp"))
    >>> func(1)
    func(param1=1, param2=2)

    You can also rename these:

    >>> class Base:
    ...     def method(self, param):
    ...         print(get_init_args(self.method, locals(), "rp", "Sub.func"))
    >>> obj = Base()
    >>> obj.method(4)
    Sub.func(param=4)
    """
    format_list = {
        "log": _format_log,
        "rp": _format_rp,
    }

    if inspect.isfunction(obj):
        obj_name = obj.__name__
    elif inspect.ismethod(obj):
        class_name = obj.__self__.__class__.__name__
        method_name = obj.__name__
        obj_name = f"{class_name}.{method_name}"
    elif inspect.isclass(obj):
        obj_name = obj.__name__
    # for `self`
    elif isinstance(obj, object):
        obj_name = obj.__class__.__name__
        obj = obj.__init__
    else:
        raise TypeError(f"Unable to parse {obj}.")

    sig = inspect.signature(obj).parameters.keys()
    ret = {"_obj_name": obj_name if rename is None else rename}
    for key in sig:
        ret[key] = str(locals[key])

    if ret_dict:
        return ret
    else:
        if format not in format_list.keys():
            raise ValueError(f"{format} format is not supported.")

        return format_list[format](ret, **kwargs)



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import logging
from pathlib import Path
import csv


class Filer:
    """A simple file handler to write in realtime.

    Parameters
    ----------
    path : str | Path
        File path. Support hierarchical folder structure.
    mode : str
        Mode of write.
    exist_ok : bool
        Whether to allow files that already exist.
    """

    def __init__(
        self,
        path: str | Path,
        mode: str = "a",
        exist_ok: bool = False,
        verbose: str | None = None,
    ) -> None:
        self.path = Path(path)
        self.mode = mode
        if self.path.exists() and not exist_ok:
            raise FileExistsError(f"File '{path}' already exists.")
        self.path.parent.mkdir(parents=True, exist_ok=True)

    def write(self, data: str) -> None:
        with open(self.path, self.mode) as f:
            f.write(data)


class CSVer(Filer):
    """Heavy development."""

    def __init__(self, path: str, mode: str = "a") -> None:
        super().__init__(path, mode)

    def write(self, data: str) -> None:
        with open(self.path, self.mode) as f:
            w = csv.writer(f)
            w.writerows(data)



import torch
from torch import nn

from .utils import Conv2dWithNorm, LinearWithNorm
from ..tools.docs import fill_doc


__all__ = ["FBCNet"]


class swish(nn.Module):
    def __init__(self):
        super(swish, self).__init__()

    def forward(self, x):
        return x * torch.sigmoid(x)


class LogVarLayer(nn.Module):
    def __init__(self, dim):
        super(LogVarLayer, self).__init__()
        self.dim = dim

    def forward(self, x):
        x = torch.clamp(x.var(dim=self.dim, keepdim=True), 1e-6, 1e6)
        return torch.log(x)


@fill_doc
class FBCNet(nn.Module):
    """FBCNet: A Multi-view Convolutional Neural Network for Brain-Computer
    Interface (FBCNet).

    FBCNet [1]_ employs a multi-view data representation followed by spatial
    filtering to extract spectro-spatially discriminative features. In FBCNet,
    a novel Variance layer is proposed to effectively aggregate the EEG time-
    domain information.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    bands : int
        The filter dimension of the input multi-view data.
    m : int
        Number of sptatial filters.
    stride : int
        Number of time windows. Must evenly divide `nTime`.

    References
    ----------
    .. [1] R. Mane et al., “FBCNet: A Multi-view Convolutional Neural Network
        for Brain-Computer Interface,” arXiv.org. Accessed: Mar. 09, 2023.
        [Online]. vailable: https://arxiv.org/abs/2104.01233v1
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        bands: int = 9,
        m: int = 32,
        stride: int = 4,
    ) -> None:
        super().__init__()
        assert not (nTime % stride), "nTime must be divisible by stride."
        self.stride = stride

        self.scb = nn.Sequential(
            Conv2dWithNorm(
                bands,
                m * bands,
                (nCh, 1),
                groups=bands,
                do_weight_norm=True,
                max_norm=2,
                padding=0,
            ),
            nn.BatchNorm2d(m * bands),
            swish(),
        )

        self.temporal_layer = LogVarLayer(dim=3)

        self.head = nn.Sequential(
            nn.Flatten(),
            LinearWithNorm(m * bands * stride, nCls, do_weight_norm=True, max_norm=0.5),
            nn.LogSoftmax(dim=1),
        )

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, bands, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        """
        x = self.scb(x)
        x = x.reshape([*x.shape[:2], self.stride, x.shape[3] // self.stride])
        x = self.temporal_layer(x)
        x = self.head(x)
        return x



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from typing import Literal

import numpy as np

from ..datasets.base import EEGData
from .base import TransformsEGD
from ..utils import get_init_args


class Pearsonr(TransformsEGD):
    """Calculate the Pearson product-moment correlation coefficient at the
    electrode channel level.

    By default, calculations are performed on ``edata`` with a shape of (
    `sample`, ..., `channel`, `time`). All data in non-`channel` dimensions
    will be transferred to the `time` dimension and flattened as values of a
    single `channel`. Finally, the Pearson product-moment correlation
    coefficient between channels is calculated and added to eegdata as a
    metric. The relationship between the correlation coefficient matrix, `R`,
    and the covariance matrix, `C`, is

    .. math:: R_{ij} = \\frac{ C_{ij} } { \\sqrt{ C_{ii} C_{jj} } }

    The values of `R` are between -1 and 1, inclusive.

    Parameters
    ----------
    prefix : str, optinoal
        Add a prefix to the metric key.
    mode : Literal["sample", "class", "all"]
        Defines sample-based calculations. Should be one of the following:

        - ``all``: Calculates the PCC value of all samples between channels
        - ``class``: Calculates the PCC value of samples of the same class
          between channels
        - ``sample``: Calculates the PCC value of each sample between channels

    Examples
    --------
    Calculate the PCC value between all samples of the channel by specifying
    the ``mode`` value:

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(8, 9, 1, 3, 10),
    ...                         label=np.random.randint(0, 2, 8))
    >>> transforms.Pearsonr(prefix="p", mode="all")(eegdata, verbose=False)
    [edata=(8, 9, 1, 3, 10), label=(8,), p_pcc_a=(3, 3)]

    or calculation based on class:

    >>> transforms.Pearsonr(mode="class")(eegdata, verbose=False)
    [edata=(8, 9, 1, 3, 10), label=(8,), p_pcc_a=(3, 3), pcc_c0=(3, 3), pcc_c1=(3, 3)]

    or calculation based on sample:

    >>> transforms.Pearsonr(mode="sample")(eegdata, verbose=False)
    [edata=(8, 9, 1, 3, 10), label=(8,), p_pcc_a=(3, 3), pcc_c0=(3, 3), pcc_c1=(3, 3), pcc_s=(8, 3, 3)]
    """

    def __init__(
        self,
        prefix: str | None = None,
        mode: Literal["sample", "class", "all"] = "sample",
    ):
        super().__init__(get_init_args(self, locals(), "rp"))
        self.key = "pcc" if prefix is None else f"{prefix}_pcc"
        self.mode = mode

    def _apply_egd(self, egd: EEGData, key: str | None):
        sp, *_, nc, _ = egd["edata"].shape
        eeg = egd["edata"].swapaxes(1, -2).reshape(sp, nc, -1)

        if self.mode == "all":
            eeg = eeg.swapaxes(0, 1).reshape(nc, -1)
            egd[f"{self.key}_a"] = np.corrcoef(eeg)
        elif self.mode == "sample":
            egd[f"{self.key}_s"] = np.array([np.corrcoef(e) for e in eeg])
        elif self.mode == "class":
            for cls in egd.cls:
                cls_idx = np.where(cls == egd["label"])
                cls_eeg = eeg[cls_idx].swapaxes(0, 1).reshape(nc, -1)
                egd[f"{self.key}_c{cls}"] = np.corrcoef(cls_eeg)
        else:
            raise ValueError(
                "Parameter mode only support `sample`, `class`, and `all`, "
                f"but got `{self.mode}`."
            )



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from abc import ABC, abstractmethod
from typing import overload, Literal

from mne.utils import verbose, logger
import test

from ..datasets.base import (
    _DataAlias,
    _DataVar,
    BaseData,
    BaseDataset,
    EEGDataset,
    EEGData,
    MultiSessEEGData,
    SplitEEGData,
)
from ..utils import unpacked, DPEEG_SEED, get_init_args, iterable_to_str


class Transforms(ABC):
    """Base class for eegdata transformation.

    The base class is callable, and the `__call__` function internally calls
    the `_apply` function according to the input data type. The allowed input
    types are eegdata or eegdataset. It also supports returning the transformed
    data of all subjects in the dataset and iteratively returning the
    transformed data of all subjects in the dataset and iteratively returning
    the transformed data of each subject. Transformation methods that inherit
    from this base class only need to override the `_apply` function.

    Attributes
    ----------
    subject : int, None
        When the input type is a dataset, if refers to the subject currently
        being operated on. Otherwise, it is None.
    """

    def __init__(self, repr: str | None = None) -> None:
        self._repr = repr

    @overload
    def __call__(self, eegdata: BaseData, verbose=None):
        """Apply data transformation to eegdata."""
        pass

    @overload
    def __call__(
        self, eegdata: BaseDataset, iter: Literal[False] = False, verbose=None
    ):
        """Apply data transformation to eegdataset."""
        pass

    @overload
    def __call__(self, eegdata: BaseDataset, iter: Literal[True] = True, verbose=None):
        """Apply data transformation to eegdataset."""
        pass

    @verbose
    def __call__(self, eegdata, iter: bool = False, verbose=None):  # type: ignore
        """Apply data transformation to eegdata or eegdataset.

        Parameters
        ----------
        eegdata : :ref:`eeg_data`, :ref:`eeg_dataset`
            Apply data transformation to eegdata or eegdataset.
        iter : bool
            Valid when the input type is eegdataset. ``False`` means directly
            returning the entire transformed dataset, ``True`` means
            iteratively returning the transformed data of each subject.

        Examples
        --------
        Allows transformation of data of type eegdata:

        >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
        ...                         label=np.random.randint(0, 3, 16))
        >>> transforms.Unsqueeze()(eegdata, verbose=False)
        [edata=(16, 1, 3, 10), label=(16,)]

        Also allows input type eegdataset:

        >>> eegdataset = dpeeg.datasets.EEGDataset([
        ...     eegdata.copy(), eegdata.copy(), eegdata.copy()
        ... ])
        >>> transforms.Squeeze()(eegdataset, iter=False, verbose=False)
        [EEGDataset:
          [eegdataset]: Subjects=3, type=EEGData
          [event_id]: None
        ]

        setting `iter` can iteratively obtain the eegdata of each subject after
        transformation:

        >>> tran = transforms.Unsqueeze()
        >>> for subject, eegdata in tran(eegdataset, iter=True, verbose=False):
        ...     print(subject, eegdata)
        0 [edata=(16, 1, 3, 10), label=(16,)]
        1 [edata=(16, 1, 3, 10), label=(16,)]
        2 [edata=(16, 1, 3, 10), label=(16,)]
        """
        if isinstance(eegdata, BaseData):
            return self._apply_wrap(eegdata, None)

        elif isinstance(eegdata, BaseDataset):
            if iter:
                return (
                    (subject, self._apply_wrap(egd, subject))
                    for subject, egd in eegdata.items()
                )
            else:
                dataset = EEGDataset(
                    event_id=eegdata.event_id, rename=eegdata._repr["_obj_name"]
                )
                for subject, egd in eegdata.items():
                    logger.info(f"Transform subject {subject}")
                    dataset[subject] = self._apply_wrap(egd, subject)
                return dataset

        else:
            raise TypeError("Unsupported eegdata type.")

    def _apply_wrap(self, eegdata: BaseData, subject: int | None) -> BaseData:
        logger.info(f"Apply {self} ...")
        self.subject = subject
        return self._apply(eegdata)

    @abstractmethod
    def _apply(self, eegdata: BaseData) -> BaseData:
        pass

    def __repr__(self) -> str:
        if self._repr:
            return self._repr
        else:
            class_name = self.__class__.__name__
            return f"{class_name} not implement attribute `self._repr`."


class TransformsEGD(Transforms):
    """This is a simple base class inherited from ``Transforms``, which only
    overrides the ``_apply`` method. If the transformation is indifferent to
    data types, one can inherit from this class and override the ``_apply_egd``
    method.
    """

    def _apply(self, eegdata: BaseData) -> BaseData:
        for egd, key in eegdata._datas():
            self._apply_egd(egd, key)
        return eegdata

    @abstractmethod
    def _apply_egd(self, egd: EEGData, key: str | None):
        pass


class Sequential(Transforms):
    """A sequential container.

    Transforms will be added to it in the order they are passed.

    Parameters
    ----------
    transforms : sequential of :doc:`../api/transforms`
        Sequential of transforms to compose.

    Examples
    --------
    If you have multiple transforms that are processed sequentiallt, you can do
    like:

    >>> transforms.Sequential(
    ...     transforms.Unsqueeze(),
    ...     transforms.Crop(2, 5),
    ... )
    >>> trans
    Sequential(
     (0): Unsqueeze(key=edata, dim=1)
     (1): Crop(tmin=2, tmax=5, include_tmax=False)
    )

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> trans(eegdata, verbose=False)
    [edata=(16, 1, 3, 3), label=(16,)]
    """

    def __init__(self, *transforms: Transforms) -> None:
        super().__init__()
        self.trans: list[Transforms] = []
        self.appends(*transforms)

    def __getitem__(self, key):
        """Return the transform of key."""
        return self.trans[key]

    def _apply_wrap(self, eegdata: BaseData, subject: int | None) -> BaseData:
        self.subject = subject
        return self._apply(eegdata)

    def _apply(self, eegdata: BaseData) -> BaseData:
        for tran in self.trans:
            eegdata = tran._apply_wrap(eegdata, self.subject)
        return eegdata

    def __repr__(self) -> str:
        s = "Sequential("
        if len(self.trans) == 0:
            return s + ")"
        else:
            for idx, tran in enumerate(self.trans):
                s += f"\n ({idx}): {tran}"
        return s + "\n)"

    def appends(self, *transforms: Transforms) -> None:
        """Append transforms to the last of composes."""
        trans = unpacked(*transforms)
        for tran in trans:
            if isinstance(tran, Sequential):
                self.trans.extend(tran.get_transforms())
            else:
                self.trans.append(tran)

    def insert(self, index: int, transform: Transforms) -> None:
        """Insert a transform at index."""
        self.trans.insert(index, transform)

    def get_transforms(self) -> list[Transforms]:
        """Return list of Transforms."""
        return self.trans


class SplitTrainTest(Transforms):
    """Split the data into training and testing sets.

    Split different types of input data. For inputs that are already
    :class:`.SplitEEGData`, no processing is done. For inputs that are
    :class:`.MultiSessEEGData`, data can be split by session or merged and then
    split. For :class:`.EEGData` inputs, data is split proportionally.

    Parameters
    ----------
    test_size : float
        The proportion of the test set. Default use stratified fashion.
    cross : bool
        ``True`` indicates that data from multiple sessions will be split into
        training and test sets, working in conjunction with ``train_sessions``
        and ``test_sessions``. ``False`` indicates that data from multiple
        sessions will be merged and then split into training and test sets,
        working with ``train_sessions``. These parameter are only effective
        when the input data type is ``MultiSessEEGData``; they are ignored for
        other types.
    train_sessions : list of str, optional
        Session data to be used as the training set.
        If ``cross=False``, ``train_sessions`` represents the sessions to be
        mixed and split (If ``None``, all session data will be used.).
        If ``cross=True``, ``train_sessions`` must be specified and represents
        the sessions to be used as the training set.
    test_sessions : list of str, optional
        Session data to be used as the test set.
        If ``cross=False``, this parameter is ignored.
        If ``cross=True``, ``test_sessions`` represents the sessions to be used
        as the test set (If ``None``, the complement of `train_sessions` will
        be used as the test set.).
    keys : list of str, optional
        The key of the eegdata value to be split. If ``None``, all data will be
        split, and it is necessary to ensure the uniformity of the data
        samples. Ignored when ``cross=True``.
    seed : int
        Random seed when splitting.

    Returns
    -------
    split_eegdata : :class:`.SplitEEGData`
        Split eegdata or dataset.

    Examples
    --------
    Split the eegdata:

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> transforms.SplitTrainTest(0.2)(eegdata, verbose=False)
    Train: [edata=(12, 3, 10), label=(12,)]
    Test : [edata=(4, 3, 10), label=(4,)]

    Split eegdata across multiple sessions:

    >>> multi_sess_eegdata = dpeeg.MultiSessEEGData(
    ...     [eegdata.copy() for _ in range(4)])
    >>> transforms.SplitTrainTest(
    ...     cross=True,
    ...     train_sessions=[f"session_{i + 1}" for i in range(2)],
    ...     test_sessions=[f"session_4"]
    ... )(multi_sess_eegdata, verbose=False)
    Train: [edata=(32, 3, 10), label=(32,)]
    Test : [edata=(16, 3, 10), label=(16,)]

    or split the merged multiple sessions eegdata:

    >>> transforms.SplitTrainTest(0.5)(multi_sess_eegdata, verbose=False)
    Train: [edata=(32, 3, 10), label=(32,)]
    Test : [edata=(32, 3, 10), label=(32,)]
    """

    def __init__(
        self,
        test_size: float = 0.25,
        cross: bool = False,
        train_sessions: list[str] | None = None,
        test_sessions: list[str] | None = None,
        keys: list[str] | None = None,
        seed: int = DPEEG_SEED,
    ) -> None:
        super().__init__(get_init_args(self, locals(), format="rp"))
        assert 0 < test_size < 1, "`test_size` should be between 0 and 1."

        self.test_size = test_size
        self.cross = cross
        self.seed = seed
        self.keys = keys
        self.train_sessions = None if train_sessions is None else set(train_sessions)
        self.test_sessions = None if test_sessions is None else set(test_sessions)

    def _apply(self, eegdata: BaseData) -> SplitEEGData:
        if isinstance(eegdata, SplitEEGData):
            return eegdata
        elif isinstance(eegdata, MultiSessEEGData):
            return self._apply_multi_sess_eegdata(eegdata)
        elif isinstance(eegdata, EEGData):
            return self._apply_eegdata(eegdata)
        else:
            raise ValueError(
                "Input should be `SplitEEGData`, `MultiSessEEGData` or `EEGData`, "
                f"but got {type(eegdata)}."
            )

    def _apply_eegdata(self, eegdata: EEGData) -> SplitEEGData:
        from sklearn.model_selection import train_test_split

        label = eegdata["label"]
        keys, values = [], []
        for key, value in eegdata.items():
            if (self.keys is None) or (key in self.keys):
                keys.append(key)
                values.append(value)

        splitting = train_test_split(
            *values, test_size=self.test_size, random_state=self.seed, stratify=label
        )

        i = 0
        train_egd, test_egd = EEGData(), EEGData()
        for key in keys:
            train_egd[key], test_egd[key] = splitting[i], splitting[i + 1]
            i += 2

        return SplitEEGData(train_egd, test_egd)

    def _merge_multi_sess_eegdata(
        self, eegdata: MultiSessEEGData, sessions: set
    ) -> EEGData:
        egd = EEGData()
        for session in sessions:
            egd.append(eegdata[session])
        return egd

    def _apply_multi_sess_eegdata(self, eegdata: MultiSessEEGData) -> SplitEEGData:
        sessions = set(eegdata.keys())

        if self.cross:
            # check `self.train_sessions`
            if self.train_sessions is None:
                raise ValueError(
                    "`train_sessions` is needed when the eegdata is "
                    "`MultiSessEEGData` and `cross=True`."
                )

            train_inter = self.train_sessions & sessions
            if len(train_inter) == 0:
                raise KeyError(
                    f"Cannot find {iterable_to_str(self.train_sessions)}, "
                    f"eegdata only contains {iterable_to_str(sessions)}."
                )
            train_compl = self.train_sessions - sessions
            if len(train_compl) != 0:
                logger.warning(
                    f"Cannot find {iterable_to_str(train_compl)}, "
                    f"only load {iterable_to_str(train_inter)} as train."
                )

            # check `self.test_sessions`
            if self.test_sessions is None:
                test_sessions = sessions - train_inter
            else:
                test_inter = self.test_sessions & sessions
                if len(test_inter) == 0:
                    test_sessions = sessions - train_inter
                    logger.warning(
                        f"Cannot find {iterable_to_str(self.test_sessions)}, "
                        f"only load {iterable_to_str(test_sessions)} as test."
                    )
                else:
                    test_sessions = test_inter
                    test_compl = self.test_sessions - sessions
                    if len(test_compl) != 0:
                        logger.warning(
                            f"Cannot find {iterable_to_str(test_compl)}, only "
                            f"load {iterable_to_str(test_sessions)} as test."
                        )

            train_eegdata = self._merge_multi_sess_eegdata(eegdata, train_inter)
            test_eegdata = self._merge_multi_sess_eegdata(eegdata, test_sessions)
            data = SplitEEGData(train_eegdata, test_eegdata)

        else:
            # check `self.train_sessions`
            if self.train_sessions is None:
                train_sessions = sessions
            else:
                train_inter = sessions & self.train_sessions
                if len(train_inter) == 0:
                    train_sessions = sessions
                    logger.warning(
                        f"Cannot find {iterable_to_str(self.train_sessions)}, "
                        f"use {iterable_to_str(train_sessions)} as mixed."
                    )
                else:
                    train_sessions = train_inter
                    train_compl = self.train_sessions - sessions
                    if len(train_compl) != 0:
                        logger.warning(
                            f"Cannot find {iterable_to_str(train_compl)}, only "
                            f"load {iterable_to_str(train_sessions)} as mixed."
                        )

            data = SplitEEGData(EEGData(), EEGData())
            for session in train_sessions:
                split = self._apply_eegdata(eegdata[session])
                data["train"].append(split["train"])
                data["test"].append(split["test"])

        return data


class ToEEGData(Transforms):
    """Convert different types of eegdata to :class:`.EEGData`.

    Merge all internal EEGData data of SplitEEGData and MultiSessEEGData
    together. This is done by calling :meth:`.EEGData.append`.

    Examples
    --------
    Convert :class:`.SplitEEGData` to :class:`.EEGData`:

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                             label=np.random.randint(0, 3, 16))
    >>> split_eegdata = dpeeg.SplitEEGData(eegdata, eegdata)
    >>> transforms.ToEEGData()(split_eegdata, verbose=False)
    [edata=(32, 3, 10), label=(32,)]

    Convert :class:`.MultiSessEEGData` to :class:`.EEGData`:

    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> multi_sess_eegdata = dpeeg.MultiSessEEGData([eegdata, eegdata])
    >>> transforms.ToEEGData()(multi_sess_eegdata, verbose=False)
    [edata=(32, 3, 10), label=(32,)]
    """

    def __init__(self) -> None:
        super().__init__("ToEEGData()")

    def _apply(self, eegdata: BaseData) -> EEGData:
        if isinstance(eegdata, EEGData):
            return eegdata
        elif isinstance(eegdata, MultiSessEEGData):
            logger.info(f" `MultiSessEEGData` to `EEGData`")
            return self._apply_multi_sess_eegdata(eegdata)
        elif isinstance(eegdata, SplitEEGData):
            logger.info(f" `SplitEEGData` to `EEGData`")
            return self._apply_split_eegdata(eegdata)
        else:
            raise ValueError(
                "Input should be `SplitEEGData`, `MultiSessEEGData` or `EEGData`, "
                f"but got {type(eegdata)}."
            )

    def _apply_split_eegdata(self, eegdata: SplitEEGData) -> EEGData:
        egd = EEGData()
        egd.append(eegdata["train"])
        egd.append(eegdata["test"])
        return egd

    def _apply_multi_sess_eegdata(self, eegdata: MultiSessEEGData) -> EEGData:
        egd = EEGData()
        for tmp_egd, _ in eegdata._datas():
            egd.append(tmp_egd)
        return egd


def merge_subjects(
    eegdata: BaseDataset,
    subjects: list[int] | set[int],
    prefix: str = "",
    ret_eegdata: bool = True,
):
    if len(prefix) != 0:
        prefix = f"{prefix}_"

    multi_sess_eegdata = MultiSessEEGData({})
    for subject in subjects:
        egd = eegdata[subject]
        ps = f"{prefix}{subject}"

        if isinstance(egd, EEGData):
            multi_sess_eegdata[ps] = egd
        elif isinstance(egd, MultiSessEEGData):
            for k, v in egd.items():
                multi_sess_eegdata[f"{ps}_{k}"] = v
        elif isinstance(egd, SplitEEGData):
            multi_sess_eegdata[f"{ps}_train"] = egd["train"]
            multi_sess_eegdata[f"{ps}_test"] = egd["test"]

    if ret_eegdata:
        return ToEEGData()._apply_multi_sess_eegdata(multi_sess_eegdata)
    else:
        return multi_sess_eegdata


@overload
def split_subjects(
    eegdata: BaseDataset,
    test_subjects: list[int] | None = None,
    ret_eegdata: Literal[True] = True,
    verbose=None,
) -> EEGData | SplitEEGData:
    pass


@overload
def split_subjects(
    eegdata: BaseDataset,
    test_subjects: list[int] | None = None,
    ret_eegdata: Literal[False] = False,
    verbose=None,
) -> MultiSessEEGData:
    pass


@verbose
def split_subjects(
    eegdata: BaseDataset,
    test_subjects: list[int] | None = None,
    ret_eegdata: bool = True,
    verbose=None,
):
    """Split the dataset by subjects.

    Splitting the dataset at the subject level is different from the
    :class:`.SplitTrainTest` transformation. The former splits the data of all
    subjects in the entire dataset (similar to cross-subject), while the latter
    splits the data of each subject. The eegdata of different subjects are
    converted through :class:`ToEEGData` when ``ret_eegdata=True``.

    Parameters
    ----------
    eegdata : :ref:`eeg_dataset`
        Input eeg dataset.
    test_subjects : list of int, optional
        The list of subjects in the test set. If ``None``, the subject data of
        the entire dataset will be merged.
    ret_eegdata : bool
        Transform the merged inter-subject data into :class:`.EEGData` type. If
        ``False``, return labeled MultiSessEEGData. For specific usage, please
        refer to the example.


    .. attention::
        Since this transformation will change the structure of the entire
        dataset, it cannot be used with :class:`Sequential`. It is often used
        at the begining or end of preprocessing the dataset.

    Examples
    --------
    Since the input is a dataset type, first define a dataset with 2 subjects.
    Here, merge the data of all subjects:

    >>> from dpeeg.datasets import EEGDataset
    >>>
    >>> eegdata = dpeeg.EEGData(edata=np.random.randn(16, 3, 10),
    ...                         label=np.random.randint(0, 3, 16))
    >>> multi_sess_eegdata = dpeeg.MultiSessEEGData([eegdata, eegdata])
    >>> eegdataset = EEGDataset([eegdata, eegdata, multi_sess_eegdata])
    >>> transforms.split_subjects(eegdataset, verbose=False)
    [edata=(64, 3, 10), label=(64,)]

    set ``train_subjects`` parameters to split the dataset:

    >>> transforms.split_subjects(eegdataset, [1, 2], verbose=False)
    Train: [edata=(32, 3, 10), label=(32,)]
    Test : [edata=(32, 3, 10), label=(32,)]

    set ``ret_eegdata`` to return the labeled MultiSessEEGData:

    >>> transforms.split_subjects(eegdataset, [3], False, verbose=False)
    {'train_1': [edata=(16, 3, 10), label=(16,)],
     'train_2': [edata=(16, 3, 10), label=(16,)],
     'test_3_session_1': [edata=(16, 3, 10), label=(16,)],
     'test_3_session_2': [edata=(16, 3, 10), label=(16,)]}
    """
    if not isinstance(eegdata, BaseDataset):
        raise TypeError(f"Input must be dataset.")

    if test_subjects is None:
        logger.info("Merge all subject eegdata.")
        return merge_subjects(eegdata, eegdata.keys(), ret_eegdata=ret_eegdata)
    else:
        subjects = set(eegdata.keys())
        te_subjects = set(test_subjects)
        inter = subjects & te_subjects
        compl = subjects - inter

        if len(inter) == 0:
            raise ValueError(
                f"Cannot find {iterable_to_str(te_subjects)} subjects, "
                f"the dataset contains {iterable_to_str(compl)} subjects."
            )

        logger.info("Split the subject eegdata.")
        if ret_eegdata:
            tr_egd = merge_subjects(eegdata, compl, ret_eegdata=True)
            te_egd = merge_subjects(eegdata, inter, ret_eegdata=True)
            ret_egd = SplitEEGData(tr_egd, te_egd)  # type: ignore
        else:
            ret_egd = MultiSessEEGData([])
            ret_egd.update(merge_subjects(eegdata, compl, "train", False))
            ret_egd.update(merge_subjects(eegdata, inter, "test", False))

        return ret_egd



from mne.channels import make_standard_montage
from mne.io import read_raw_edf

from .base import DATA_PATH, EpochsDataset
from .download import data_dl
from ..utils import get_init_args
from ..tools.docs import fill_doc


URL = "https://web.gin.g-node.org/"


@fill_doc
class HighGamma(EpochsDataset):
    """High-gamma dataset described in Schirrmeister et al. 2017.

    .. admonition:: Dataset summary

        ====== ====== ======= ===== ======== ======
        Subj   Chan   Time    Cls   Freq     Sess
        ====== ====== ======= ===== ======== ======
        14     128    4.0 s   4     500 Hz   2
        ====== ====== ======= ===== ======== ======

    This data set consists of EEG data from 14 subjects of a study published in
    [1]_. High-Gamma Dataset is a 128-electrode dataset (of which we later only
    use 44 sensors covering the motor cortex, obtained from 14 healthy subjects
    (6 female, 2 left-handed, age 27.2 ± 3.6 (mean ± std)) with roughly 1000
    (963.1 ± 150.9, mean ± std) four-second trials of executed movements
    divided into 13 runs per subject. The four classes of movements were
    movements of either the left hand, the right hand, both feet, and rest (no
    movement, but same type of visual cue as for the other classes). The
    training set consists of the approx. 880 trials of all runs except the last
    two runs (as session 1), the test set of the approx. 160 trials of the last
    2 runs (as session 2). This dataset was acquired in an EEG lab optimized
    for non-invasive detection of high- frequency movement-related EEG
    components.

    Depending on the direction of a gray arrow that was shown on black back-
    ground, the subjects had to repetitively clench their toes (downward arrow),
    perform sequential finger-tapping of their left (leftward arrow) or right
    (rightward arrow) hand, or relax (upward arrow). The movements were selected
    to require little proximal muscular activity while still being complex
    enough to keep subjects in- volved. Within the 4-s trials, the subjects
    performed the repetitive movements at their own pace, which had to be
    maintained as long as the arrow was showing. Per run, 80 arrows were
    displayed for 4 s each, with 3 to 4 s of continuous random inter-trial
    interval. The order of presentation was pseudo-randomized, with all four
    arrows being shown every four trials. Ideally 13 runs were performed to
    collect 260 trials of each movement and rest. The stimuli were presented
    and the data recorded with BCI2000. The experiment was approved by the
    ethical committee of the University of Freiburg.

    References
    ----------
    .. [1] Schirrmeister, Robin Tibor, et al. "Deep learning with convolutional
        neural networks for EEG decoding and visualization." Human brain
        mapping 38.11 (2017): 5391-5420.

    Parameters
    ----------
    %(subjects)s
    %(epochs_tmin_tmax)s
    %(baseline)s
    %(picks)s
    %(resample)s
    %(rename)s
    """

    def __init__(
        self,
        subjects: list[int] | None = None,
        tmin: float = 0,
        tmax: float | None = None,
        baseline: tuple[int, int] | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        rename: str | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), rename=rename, ret_dict=True),
            subject_list=list(range(1, 15)),
            interval=[0.0, 4.0],
            event_id={"right_hand": 1, "left_hand": 2, "rest": 3, "feet": 4},
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            baseline=baseline,
            picks=picks,
            resample=resample,
        )
        self._data_url = f"{URL}robintibor/high-gamma-dataset/raw/master/data/"
        self._data_path = DATA_PATH / "high_gamma"

    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        sessions = {}
        montage = make_standard_montage("standard_1005")
        for i, r in enumerate(["train", "test"], start=1):
            filename = data_dl(
                f"{self._data_url}{r}/{subject:d}.edf",
                self._data_path / r,
                force_update=False,
            )
            raw = read_raw_edf(
                filename, infer_types=True, preload=True, verbose=verbose
            )
            raw.set_montage(montage)
            sessions.update({f"session_{i}": {"run_1": raw}})

        return sessions



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from pathlib import Path

from pooch import file_hash, retrieve
from pooch.downloaders import choose_downloader


def data_dl(
    url: str | Path,
    path: str | Path,
    force_update: bool = False,
):
    """Download file from url to specified path.

    Parameters
    ----------
    url : str
        Path to remote location of data.
    path : str
        Location of where to look for the data storing location. Default is
        `~/dpeeg/datasets`. If the dataset is not found under the given path,
        the data will be automatically downloaded to the specified folder.
    force_update : bool
        Force update of the dataset even if a local copy exists.

    Returns
    -------
    path : str
        Local path to the given data file.
    """
    destination = Path(path)

    downloader = choose_downloader(url, progressbar=True)
    if type(downloader).__name__ in ["HTTPDownloader", "DOIDownloader"]:
        downloader.kwargs.setdefault("verify", False)

    # Fetch the file
    if not destination.is_file() or force_update:
        if destination.is_file():
            destination.unlink()
        destination.parent.mkdir(parents=True, exist_ok=True)
        known_hash = None
    else:
        known_hash = file_hash(str(destination))
    dlpath = retrieve(
        url,
        known_hash,
        fname=Path(url).name,
        path=str(destination),
        progressbar=True,
        downloader=downloader,
    )
    return dlpath



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import platform

import torch
from torch import Tensor
from torch.nn import Module


def get_device(nGPU: int = 0) -> torch.device:
    """Get the available devices.

    Parameters
    ----------
    nGPU : int
        GPU number to use.

    Returns
    -------
    device : torch.device
        Available device.
    """
    if not torch.cuda.is_available():
        dev = torch.device("cpu")
    else:
        if (nGPU > torch.cuda.device_count() - 1) or (nGPU < 0):
            raise ValueError(f"GPU: {nGPU} does not exit.")
        dev = torch.device(f"cuda:{nGPU}")
    return dev


def get_device_name(device: torch.device) -> str:
    if device.type == "cuda":
        return torch.cuda.get_device_name(device)
    elif device.type == "cpu":
        return platform.processor() or "Unknown CPU"
    else:
        return "Unsupported device type"


def model_predict(
    model: Module,
    data: Tensor,
    event_id: dict[str, int] | None = None,
    no_grad: bool = True,
) -> tuple[Tensor, list[str]] | tuple[Tensor, None]:
    """Compute model predictions.

    Parameters
    ----------
    model : Module
        Inherit Module and should define the forward method. The first
        parameter returned by model forward propagation is the prediction.
    data : Tensor (N, ...)
        Data to be predicted.
    event_id : dict, optional
        A dictionary containing label and corresponding id.
    no_grad : bool
        Whether to turn off autograd calculation graph recording.

    Returns
    -------
    preds : Tensor (N, ...)
        Prediction corresponding to the input data.
    labels : list of str
        If event_id not None, corresponding predicted label will be returned.
    """
    model.eval()
    if no_grad:
        with torch.no_grad():
            out = model(data)
    else:
        out = model(data)

    out = out[0] if isinstance(out, tuple) else out
    preds = torch.argmax(out, dim=1).detach()

    if event_id:
        reversed_event_id = {v: k for k, v in event_id.items()}
        labels = [reversed_event_id[pred] for pred in preds.tolist()]
        return preds, labels
    else:
        return preds, None


def model_depth(model: Module) -> int:
    """Compute the maximum depth of the model."""
    names = dict(model.named_modules()).keys()
    name_depth = [len(name.split(".")) for name in names]
    max_depth = max(name_depth)
    return max_depth



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import torch
import torch.nn as nn


class Conv2dWithNorm(nn.Conv2d):
    def __init__(self, *args, do_weight_norm=True, max_norm=1.0, p=2, **kwargs):
        super().__init__(*args, **kwargs)
        self.p = p
        self.max_norm = max_norm
        self.do_weight_norm = do_weight_norm

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.do_weight_norm:
            self.weight.data = torch.renorm(self.weight.data, self.p, 0, self.max_norm)
        return super().forward(input)

    def __repr__(self):
        repr = super().__repr__()
        if self.do_weight_norm:
            last_bracket_index = repr.rfind(")")
            self_repr = f", max_norm={self.max_norm}, p={self.p}"
            repr = repr[:last_bracket_index] + self_repr + ")"
        return repr


class LinearWithNorm(nn.Linear):
    def __init__(self, *args, do_weight_norm=True, max_norm=1.0, p=2, **kwargs):
        super().__init__(*args, **kwargs)
        self.p = p
        self.max_norm = max_norm
        self.do_weight_norm = do_weight_norm

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.do_weight_norm:
            self.weight.data = torch.renorm(self.weight.data, self.p, 0, self.max_norm)
        return super().forward(input)

    def __repr__(self):
        repr = super().__repr__()
        if self.do_weight_norm:
            last_bracket_index = repr.rfind(")")
            self_repr = f", max_norm={self.max_norm}, p={self.p}"
            repr = repr[:last_bracket_index] + self_repr + ")"
        return repr



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from abc import abstractmethod

from numpy import ndarray
from mne.utils import verbose, logger

from .base import Transforms
from ..datasets.base import EEGData, _BaseData, SplitEEGData
from ..utils import get_init_args
from ..tools.docs import fill_doc
from .functions import (
    z_score_norm,
    min_max_norm,
)


class Normalization(Transforms):
    """Data normalization base class.

    Normalize the data, with default normalization applied to the `edata`.
    Please verify the validity of the data.

    Attributes
    ----------
    params : dict
        Saves intermediate variables of calculations.
    """

    def __init__(self, repr: str | None = None, train_for_test: bool = False) -> None:
        super().__init__(repr)
        self.train_for_test = train_for_test
        self.params = {}

    def _apply(self, eegdata: _BaseData) -> _BaseData:
        if (not isinstance(eegdata, SplitEEGData)) and self.train_for_test:
            raise TypeError(
                "The input must have been split when `train_for_test` is True."
            )
        else:
            for egd, mode in eegdata._datas():
                self._apply_norm(egd, mode)

        return eegdata

    @abstractmethod
    def _apply_norm(self, egd: EEGData, mode: str):
        pass


@fill_doc
class ZscoreNorm(Normalization):
    r"""Z-score normalization per subject.

    By default, the EEG data (``edata``) of eegdata are normalized.

    .. math::

       \mathbf{z} = \frac{\mathbf{x} - \mu}{\sqrt{\sigma^{2}}}

    where :math:`\mathbf{x}` and :math:`\mathbf{z}` denote the input data and
    the output of normalization, respectively. :math:`\mu` and :math:`\sigma^2`
    represent the mean and variance values of the sample.

    Parameters
    ----------
    %(train_for_test)s
    mean : ndarray, optional
        The mean used in the normalization process. If None, use the statistics
        of the current sample for normalization. Ignored when ``train_for_test``
        is True.
    std : ndarray, optional
        The standard deviation used in the normalization process. If None, use
        the statistics of the current sample for normalization. Ignored when
        ``train_for_test`` is True.
    dim : int, optional
        The dimension to normalize. Usually, -1 for channels and -2 for time
        points. If None, normalize at the sample level.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.
    """

    def __init__(
        self,
        train_for_test: bool = False,
        mean: ndarray | None = None,
        std: ndarray | None = None,
        dim: int | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), format="rp"),
            train_for_test=train_for_test,
        )
        self.mean = mean
        self.std = std
        self.dim = dim

    def _apply_norm(self, egd: EEGData, mode: str):
        if self.train_for_test:
            if mode == "train":
                egd["edata"], mean, std = z_score_norm(
                    egd["edata"], dim=self.dim, ret=True
                )
                self.params[self.subject] = {"mean": mean, "std": std}
            else:
                egd["edata"] = z_score_norm(
                    egd["edata"],
                    mean=self.params[self.subject]["mean"],
                    std=self.params[self.subject]["std"],
                    dim=self.dim,
                    ret=False,
                )
        else:
            egd["edata"], mean, std = z_score_norm(
                egd["edata"], self.mean, self.std, self.dim, True
            )
            self.params.update({"mean": mean, "std": std})


@fill_doc
class MinMaxNorm(Normalization):
    r"""Min-max normalization per subject.

    By default, the EEG data (``edata``) of eegdata are normalized.

    .. math::

       \mathbf{z} =
       \frac{\mathbf{x}-\mathbf{x}_{min}}{\mathbf{x}_{max}-\mathbf{x}_{min}}

    where :math:`\mathbf{x}` and :math:`\mathbf{z}` denote the input data and
    the output of normalization, respectively. :math:`\mathbf{x}_{max}` and
    :math:`\mathbf{x}_{min}` represent the maximum and minimum values of the
    sample.

    Parameters
    ----------
    %(train_for_test)s
    min : ndarray, optional
        The minimum used in the normalization process. If None, use the
        statistics of the current sample for normalization. Ignored when
        ``train_for_test`` is True.
    max : ndarray, optional
        The maximum used in the normalization process. If None, use the
        statistics of the current smaple for normalization. Ignored when
        ``train_for_test`` is True.
    dim : int, optional
        The dimension to normalize. Usually, -1 for channels and -2 for time
        points. If None, normalize at the sample level.

    Returns
    -------
    data : eegdata or dataset
        Transformed eegdata.
    """

    def __init__(
        self,
        train_for_test: bool = False,
        min: ndarray | None = None,
        max: ndarray | None = None,
        dim: int | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), format="rp"),
            train_for_test=train_for_test,
        )
        self.train_for_test = train_for_test
        self.min = min
        self.max = max
        self.dim = dim

    def _apply_norm(self, egd: EEGData, mode: str):
        if self.train_for_test:
            if mode == "train":
                egd["edata"], min, max = min_max_norm(
                    egd["edata"], dim=self.dim, ret=True
                )
                self.params[self.subject] = {"min": min, "max": max}
            else:
                egd["edata"] = min_max_norm(
                    egd["edata"],
                    min=self.params[self.subject]["min"],
                    max=self.params[self.subject]["max"],
                    dim=self.dim,
                    ret=False,
                )
        else:
            egd["edata"], min, max = min_max_norm(
                egd["edata"], self.min, self.max, self.dim, True
            )
            self.params.update({"min": min, "max": max})



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops.layers.torch import Rearrange

from ..tools.docs import fill_doc


__all__ = ["MSVTNet", "JointCrossEntoryLoss"]


class TSConv(nn.Sequential):
    def __init__(self, nCh, F, C1, C2, D, P1, P2, Pc) -> None:
        super().__init__(
            nn.Conv2d(1, F, (1, C1), padding="same", bias=False),
            nn.BatchNorm2d(F),
            nn.Conv2d(F, F * D, (nCh, 1), groups=F, bias=False),
            nn.BatchNorm2d(F * D),
            nn.ELU(),
            nn.AvgPool2d((1, P1)),
            nn.Dropout(Pc),
            nn.Conv2d(F * D, F * D, (1, C2), padding="same", groups=F * D, bias=False),
            nn.BatchNorm2d(F * D),
            nn.ELU(),
            nn.AvgPool2d((1, P2)),
            nn.Dropout(Pc),
        )


class PositionalEncoding(nn.Module):
    def __init__(self, seq_len, d_model) -> None:
        super().__init__()
        self.seq_len = seq_len
        self.d_model = d_model
        self.pe = nn.Parameter(torch.zeros(1, seq_len, d_model))

    def forward(self, x):
        x += self.pe
        return x


class Transformer(nn.Module):
    def __init__(
        self,
        seq_len,
        d_model,
        nhead,
        ff_ratio,
        Pt=0.5,
        num_layers=4,
    ) -> None:
        super().__init__()
        self.cls_embedding = nn.Parameter(torch.zeros(1, 1, d_model))
        self.pos_embedding = PositionalEncoding(seq_len + 1, d_model)

        dim_ff = d_model * ff_ratio
        self.dropout = nn.Dropout(Pt)
        self.trans = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(
                d_model, nhead, dim_ff, Pt, batch_first=True, norm_first=True
            ),
            num_layers,
            norm=nn.LayerNorm(d_model),
        )

    def forward(self, x):
        b = x.shape[0]
        x = torch.cat((self.cls_embedding.expand(b, -1, -1), x), dim=1)
        x = self.pos_embedding(x)
        x = self.dropout(x)
        return self.trans(x)[:, 0]


class ClsHead(nn.Sequential):
    def __init__(self, linear_in, nCls):
        super().__init__(nn.Flatten(), nn.Linear(linear_in, nCls), nn.LogSoftmax(dim=1))


@fill_doc
class MSVTNet(nn.Module):
    """MSVTNet: Multi-Scale Vision Transformer Neural Network for EEG-Based
    Motor Imagery Decoding (MSVTNet).

    MSVTNet [1]_ effectively integrates the strengths of convolutional neural
    networks (CNNs) in extracting local features with the global feature
    extraction capabilities of Transformers. Specifically, to optimize
    classification features, a multi-branch CNN with different scales is
    designed to capture local spatiotemporal features, along with a Transformer
    to jointly model global and local spatiotemporal correlations features.
    Additionally, auxiliary branch loss (ABL) is leveraged for intermediate
    supervision, ensuring effective integration of CNNs and Transformers.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    F : list of int
        Number of temporal filters per branch.
    C1 : list of int
        The convolution kernel size of each branch temporal filter.
    C2 : int
        Depthwise convolution kernel size.
    D : int
        Depth of depthwise convolution.
    P1 : float
        The first pooling kernel size.
    P2 : float
        The second pooling kernel size.
    Pc : float
        Dropout rate of multi-branch convolutional module.
    nhead : int
        Number of multi-head attention.
    ff_ratio : int
        The expansion factor of the fully connected feed-forward layer.
    Pt : float
        Dropout rate of transformer encoder.
    layers : int
        Number of transformer encoder layers.
    b_preds : bool
        If ``True``, return the prediction for each branch.

    References
    ----------
    .. [1] K. Liu et al., "MSVTNet: Multi-Scale Vision Transformer Neural
        Network for EEG-Based Motor Imagery Decoding," in IEEE Journal of
        Biomedical and Health Informatics, doi: 10.1109/JBHI.2024.3450753.
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        F: list[int] = [9, 9, 9, 9],
        C1: list[int] = [15, 31, 63, 125],
        C2: int = 15,
        D: int = 2,
        P1: int = 8,
        P2: int = 7,
        Pc: float = 0.3,
        nhead: int = 8,
        ff_ratio: int = 1,
        Pt: float = 0.5,
        layers: int = 2,
        b_preds: bool = True,
    ) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime
        self.b_preds = b_preds
        assert len(F) == len(C1), "The length of F and C1 should be equal."

        self.mstsconv = nn.ModuleList(
            [
                nn.Sequential(
                    TSConv(nCh, F[b], C1[b], C2, D, P1, P2, Pc),
                    Rearrange("b d 1 t -> b t d"),
                )
                for b in range(len(F))
            ]
        )
        branch_linear_in = self._forward_flatten(cat=False)
        self.branch_head = nn.ModuleList(
            [ClsHead(branch_linear_in[b].shape[1], nCls) for b in range(len(F))]
        )

        seq_len, d_model = self._forward_mstsconv().shape[1:3]  # type: ignore
        self.transformer = Transformer(seq_len, d_model, nhead, ff_ratio, Pt, layers)

        linear_in = self._forward_flatten().shape[1]  # type: ignore
        self.last_head = ClsHead(linear_in, nCls)

    def _forward_mstsconv(self, cat=True):
        x = torch.randn(1, 1, self.nCh, self.nTime)
        x = [tsconv(x) for tsconv in self.mstsconv]
        if cat:
            x = torch.cat(x, dim=2)
        return x

    def _forward_flatten(self, cat=True):
        x = self._forward_mstsconv(cat)
        if cat:
            x = self.transformer(x)
            x = x.flatten(start_dim=1, end_dim=-1)
        else:
            x = [_.flatten(start_dim=1, end_dim=-1) for _ in x]
        return x

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, bands, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        branch_cls_prob : list of Tensor
            If ``b_preds=True``, return the class prediction probability for
            each branch.
        """
        x = [tsconv(x) for tsconv in self.mstsconv]
        bx = [branch(x[idx]) for idx, branch in enumerate(self.branch_head)]
        x = torch.cat(x, dim=2)
        x = self.transformer(x)
        x = self.last_head(x)
        if self.b_preds:
            return x, bx
        else:
            return x


class JointCrossEntoryLoss(nn.Module):
    r"""Auxiliary branch loss.

    The parameters of MSVTNet are learned under the supervision of the
    auxiliary branch loss and model prediction loss:

    .. math::

       \mathcal{L}=\lambda\mathcal{L}_c+(1-\lambda)\sum_{b=1}^{B}\mathcal{L}_b

       \mathcal{L}_{c/b}=\mathrm{Cross Entropy Loss}(\hat{y})

    where :math:`\lambda\in(0, 1]` is the ratio factor for intermediate
    supervision of the model.

    Parameters
    ----------
    lamd : float
        Ratio factor of ABL.
    """

    def __init__(self, lamd: float = 0.6) -> None:
        super().__init__()
        self.lamd = lamd

    def forward(self, out, label):
        """Forward pass function that processes the model and branch prediction
        probabilities.

        Parameters
        ----------
        out : tuple of Tensor
            Models and branch prediction probabilities.
        label : Tensor
            True label.

        Returns
        -------
        loss : Tensor
            Loss with gradient.
        """
        end_out = out[0]
        branch_out = out[1]
        end_loss = F.nll_loss(end_out, label)
        branch_loss = [F.nll_loss(out, label).unsqueeze(0) for out in branch_out]
        branch_loss = torch.cat(branch_loss)
        loss = self.lamd * end_loss + (1 - self.lamd) * torch.sum(branch_loss)
        return loss



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

import os
from abc import abstractmethod, ABC
from copy import deepcopy

import torch
import torch.nn as nn
from torch import optim
from torch import Tensor
from torch.nn import Module
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.backends import cudnn
from torchinfo import summary
from typing import Literal, Type
from torch.utils.tensorboard.writer import SummaryWriter
from torch.utils.data import TensorDataset, DataLoader, ConcatDataset
from torchmetrics.functional.classification.accuracy import multiclass_accuracy
from torchmetrics.aggregation import MeanMetric, CatMetric

from .base import Trainer
from ..utils import DPEEG_SEED
from ..transforms.functions import to_tensor
from .stopcriteria import ComposeStopCriteria
from .utils import get_device, get_device_name, model_depth
from ..utils import _set_torch_seed, mapping_to_str
from ..datasets.base import EEGData


cudnn.benchmark = False
cudnn.deterministic = True


class BaseClassifier(Trainer, ABC):
    """Classification model trainer.

    Generate a trainer to test the performance of the same network on
    different datasets.
    """

    def __init__(
        self,
        model: Module,
        loss_fn: str | Type[Module] = "NLLLoss",
        loss_fn_args: dict | None = None,
        optimizer: str | Type[Optimizer] = "Adam",
        optimizer_args: dict | None = None,
        lr: float = 1e-3,
        lr_sch: str | Type[LRScheduler] | None = None,
        lr_sch_args: dict | None = None,
        grad_acc: int = 1,
        batch_size: int = 32,
        nGPU: int = 0,
        seed: int = DPEEG_SEED,
        keep_data_gpu: bool = True,
        depth: int | None = None,
        data_size: tuple | list | None = None,
        verbose: int | str = "INFO",
    ) -> None:
        super().__init__(model, verbose)

        # init trainer
        self.device = get_device(nGPU)
        self.logger.info(
            f"Model will be trained on {self.device} "
            f"({get_device_name(self.device)})"
        )
        self.model.to(self.device)

        _set_torch_seed(seed)
        self.logger.info(f"Set torch random seed = {seed}")

        # summarize model structure
        self.model_arch = str(model) + "\n"
        depth = model_depth(self.model) if depth is None else depth
        self.model_arch += str(summary(model, data_size, depth=depth))
        self.logger.info(self.model_arch)

        # the type of optimizer, etc. selected
        self.loss_fn_type = loss_fn
        self.optimizer_type = optimizer
        self.lr_sch_type = lr_sch

        # save additional parameters
        self.loss_fn_args = loss_fn_args if loss_fn_args else {}
        self.optimizer_args = optimizer_args if optimizer_args else {}
        self.lr_sch_args = lr_sch_args if lr_sch_args else {}
        # --- others
        self.optimizer_args["lr"] = lr
        self.grad_acc = grad_acc
        self.batch_size = batch_size
        self.seed = seed
        self.keep_data_gpu = keep_data_gpu
        self.verbose = verbose

        # set experimental details
        self.train_details = {
            "type": self.__class__.__name__,
            "train_param": {
                "seed": seed,
                "loss_fn": str(loss_fn),
                "loss_fn_args": loss_fn_args,
                "optimizer": str(optimizer),
                "optimizer_args": optimizer_args,
                "lr_sch": str(lr_sch),
                "lr_sch_args": lr_sch_args,
                "batch_size": batch_size,
                "grad_acc": grad_acc,
            },
            "orig_model_param": deepcopy(self.model.state_dict()),
        }

    def fit_epoch(self, train_loader: DataLoader) -> None:
        """Fit one epoch to train model.

        Parameters
        ----------
        train_loader : DataLoader
            DataLoader used for training.
        """
        # set the model in training mode
        self.model.train()

        # iterater over all the data
        with torch.enable_grad():
            for idx, (data, label) in enumerate(train_loader):
                data, label = data.to(self.device), label.to(self.device)
                out = self.model(data)
                loss = self.loss_fn(out, label)
                loss.backward()
                # gradient accumulation
                if (idx + 1) % self.grad_acc == 0:
                    # 1 - update parameters
                    self.optimizer.step()
                    # 2 - zero the parameter gradients
                    self.optimizer.zero_grad()
            # update lr
            # Note: Learning rate scheduling should be applied after optimizer’s update
            if self.lr_sch:
                self.lr_sch.step()

    def predict(
        self,
        data_loader: DataLoader,
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Predict the class of the input data.

        Parameters
        ----------
        data_loader : DataLoader
            Dataset used for prediction.

        Returns
        -------
        preds : Tensor
            Predicted labels, as returned by a classifier.
        target : Tensor
            Ground truth (correct) labels.
        loss : Tensor
            Average loss.
        """
        # set the model in the eval mode
        self.model.eval()

        loss_sum = MeanMetric()
        preds, target = CatMetric(), CatMetric()

        # iterate over all the data
        with torch.no_grad():
            for data, label in data_loader:
                data, label = data.to(self.device), label.to(self.device)
                out = self.model(data)
                loss = self.loss_fn(out, label)
                loss_sum.update(loss.item(), data.size(0))
                # convert the output of soft-max to class label
                # save preds and actual label
                out = out[0] if isinstance(out, tuple) else out
                preds.update(torch.argmax(out, dim=1).detach().cpu())
                target.update(label.cpu())
        return preds.compute(), target.compute(), loss_sum.compute()

    def data_loader(self, *datasets: EEGData) -> DataLoader:
        """Wrap multiple sets of data and labels and return DataLoader.

        Parameters
        ----------
        datasets : sequence of EEGData
            Sequence of EEGData. Multiple EEGData will be concatenated.
        """
        if len(datasets) == 0:
            raise ValueError("At least one dataset required as input.")

        # dataset wrapping tensors
        td = []
        for dataset in datasets:
            data, label = to_tensor(dataset["edata"], dataset["label"])
            if self.keep_data_gpu:
                data, label = data.to(self.device), label.to(self.device)
            td.append(TensorDataset(data, label))
        td = ConcatDataset(td)

        return DataLoader(td, self.batch_size, True)

    def _reset_fitter(self, log_dir: str) -> tuple[str, SummaryWriter]:
        """Reset the relevant parameters of the fitter.

        Reset the model's training parameters, learning rate schedule and
        optimizer etc. to their initialized state.

        Parameters
        ----------
        log_dir : str
            Directory location (support hierarchical folder structure) to save
            training log.

        Returns
        -------
        str, SummaryWriter:
            Return the absolute file path and new SummaryWriter object manager
            for the fitter.
        """
        self._reset_logger(log_dir)

        # reset parameters of nn.Moudle
        self.model.load_state_dict(self.train_details["orig_model_param"])

        # create loss function
        if isinstance(self.loss_fn_type, str):
            self.loss_fn = getattr(nn, self.loss_fn_type)(**self.loss_fn_args)
        else:
            self.loss_fn = self.loss_fn_type(**self.loss_fn_args)

        # create optimizer
        if isinstance(self.optimizer_type, str):
            self.optimizer = getattr(optim, self.optimizer_type)(
                self.model.parameters(), **self.optimizer_args
            )
        elif issubclass(self.optimizer_type, Optimizer):
            self.optimizer = self.optimizer_type(
                self.model.parameters(), self.optimizer_args
            )
        else:
            raise TypeError(
                f"Optimizer type ({type(self.optimizer_type)}) is not supported."
            )

        # create lr_scheduler
        if self.lr_sch_type is None:
            self.lr_sch = None
        elif isinstance(self.lr_sch_type, str):
            self.lr_sch = getattr(optim.lr_scheduler, self.lr_sch_type)(
                self.optimizer, **self.lr_sch_args
            )
        elif issubclass(self.lr_sch_type, LRScheduler):
            self.lr_sch = self.lr_sch_type(self.optimizer, **self.lr_sch_args)
        else:
            raise TypeError(f"lr_sch type ({type(self.lr_sch_type)}) is not supported.")

        # create log writer
        log_dir = os.path.abspath(log_dir)
        writer = SummaryWriter(log_dir)

        return log_dir, writer

    def get_acc(self, preds: Tensor, target: Tensor, ncls: int) -> Tensor:
        """Easy for program to caculate the accuarcy."""
        return multiclass_accuracy(preds, target, ncls, "micro")

    def __repr__(self) -> str:
        s = "[Model architecture]:\n" + self.model_arch + "\n"

        s += f"[Loss function]: {self.loss_fn_type}"
        if self.loss_fn_args:
            s += f"({mapping_to_str(self.loss_fn_args)})\n"
        else:
            s += "\n"

        s += f"[Optimizer]: {self.optimizer_type}"
        if self.optimizer_args:
            s += f"({mapping_to_str(self.optimizer_args)})\n"
        else:
            s += "\n"

        if self.lr_sch_type:
            s += f"[Lr scheduler]: {self.lr_sch_type}"
            if self.lr_sch_args:
                s += f"({mapping_to_str(self.lr_sch_args)})\n"
            else:
                s += "\n"

        s += f"[Grad Acc]: {self.grad_acc}\n"
        s += f"[Batch Size]: {self.batch_size}\n"
        s += f"[Seed]: {self.seed}\n"

        return s


class Classifier(BaseClassifier):
    """Classifier training.

    During different training model processes, early stopping mechanisms can be
    executed using the training set (validation set not required) to select the
    model.

    Parameters
    ----------
    model : Module
        Inherit Module and should define the forward method. The first
        parameter returned by model forward propagation is the prediction.
    loss_fn : str, Type[Module]
        Name of the loss function from `torch.nn` which will be used for
        training. If Module, means using a custom loss function.
        Note: custom loss_fn is a class (not an instance), and its
        initialization list is `(**loss_fn_args)`.
    loss_fn_args : dict, optional
        Additional arguments to be passed to the loss function.
    optimizer : str, Type[Optimizer]
        Name of the optimization function from torch.optim which will be used
        for training. If Optimizer, means using a custom optimizer.
        Note: custom optimizer is a class (not an instance), and its
        initialization list is `(model, lr=lr, **optimizer_args)`.
    optimizer_args : dict, optional
        Additional arguments to be passed to the optimization function.
    lr : float
        Learning rate.
    lr_sch : str, Type[LRScheduler], optional
        Name of the learning scheduler from `torch.optim.lr_scheduler` which
        will be used for training. If LRScheduler, means using a custom
        learning scheduler.
        Note: custom learning scheduler is a class (not an instance), and its
        initialization list is `(optimizer, **lr_sch_args)`.
    lr_sch_args : dict, optional
        Additional arguments to be passed to the lr_scheduler function.
    grad_acc : int
        Aradient accumulation.
    batch_size : int
        Mini-batch size.
    max_epochs : int
        Maximum number of epochs in training.
    no_increase_epochs : int
        Maximum number of consecutive epochs when the accuracy or loss of the
        training set has no relative improvement.
    var_check : str
        The best value (train_inacc/train_loss) to check while determining the
        best model which will be used to evaluate its performance on the test.
    load_best_state : bool
        If `True`, the best model parameters will be used for evaluation.
    nGPU : int
        Select the gpu id to train. If the GPU is not available then the CPU is
        used.
    seed : int
        Select random seed for review.
    keep_data_gpu : bool
        Keep the dataset on the GPU to avoid the time consumption of data
        migration. Please adjust according to the personal GPU memory.
    data_size : tuple, list, optional
        Output the structure of the network model according to the input
        dimension if the `data_size` is given.
    depth : int, optional
        Depth of nested layers to display.
    verbose : int, str
        The log level of console. Default is INFO. Mainly used for debug.
    """

    def __init__(
        self,
        model: Module,
        loss_fn: str | Type[Module] = "NLLLoss",
        loss_fn_args: dict | None = None,
        optimizer: str | Type[Optimizer] = "Adam",
        optimizer_args: dict | None = None,
        lr: float = 0.001,
        lr_sch: str | Type[LRScheduler] | None = None,
        lr_sch_args: dict | None = None,
        grad_acc: int = 1,
        batch_size: int = 32,
        max_epochs: int = 1000,
        no_increase_epochs: int = 50,
        var_check: Literal["train_loss", "train_inacc"] = "train_loss",
        load_best_state: bool = True,
        nGPU: int = 0,
        seed: int = DPEEG_SEED,
        keep_data_gpu: bool = True,
        depth: int | None = None,
        data_size: tuple | list | None = None,
        verbose: int | str = "INFO",
    ) -> None:
        super().__init__(
            model,
            loss_fn,
            loss_fn_args,
            optimizer,
            optimizer_args,
            lr,
            lr_sch,
            lr_sch_args,
            grad_acc,
            batch_size,
            nGPU,
            seed,
            keep_data_gpu,
            depth,
            data_size,
            verbose,
        )
        self.max_epochs = max_epochs
        self.no_increase_epochs = no_increase_epochs
        self.var_check = var_check
        self.load_best_state = load_best_state
        self.train_details.update(
            {
                "train_param": {
                    "max_epochs": max_epochs,
                    "no_increase_epochs": no_increase_epochs,
                    "var_check": var_check,
                    "load_best_state": load_best_state,
                }
            }
        )

    def fit(
        self,
        trainset: EEGData,
        testset: EEGData,
        log_dir: str,
    ) -> dict[str, dict[str, Tensor]]:
        """Fit the model.

        Parameters
        ----------
        trainset : EEGData
            Dataset used for training.
        testset : EEGData
            Dataset used to evaluate the model.
        log_dir : str
            The path to save the training log.

        Returns
        -------
        result : dict
            Returns the training set and test set results (including true
            labels, predicted labels and accuracy).
        """
        log_dir, writer = self._reset_fitter(log_dir)

        # check the best model
        best_var = float("inf")
        best_model_param = deepcopy(self.model.state_dict())

        # initialize dataloader
        self.trainset = trainset
        self.testset = testset
        train_loader = self.data_loader(trainset)
        test_loader = self.data_loader(testset)
        train_ncls = trainset.ncls
        test_ncls = testset.ncls

        # start the training
        self.timer.start()
        self.logger.info(f"[Training...] - [{self.timer.ctime()}]")
        self.logger.info(f"[Train/Test] - [{trainset.trials()}/{testset.trials()}]")

        stopcri = ComposeStopCriteria(
            {
                "Or": {
                    "cri1": {
                        "MaxEpoch": {"max_epochs": self.max_epochs, "var_name": "epoch"}
                    },
                    "cri2": {
                        "NoDecrease": {
                            "num_epochs": self.no_increase_epochs,
                            "var_name": self.var_check,
                        }
                    },
                }
            }
        )
        monitors = {"epoch": 0, "train_loss": float("inf"), "train_inacc": 1}

        while not stopcri(monitors):

            # train one epoch
            self.fit_epoch(train_loader)
            monitors["epoch"] += 1

            # evaluate the training and validation accuracy
            train_preds, train_target, train_loss = self.predict(train_loader)
            train_acc = self.get_acc(train_preds, train_target, train_ncls)
            monitors["train_inacc"] = 1 - train_acc
            monitors["train_loss"] = train_loss

            # store loss and acc
            writer.add_scalars(
                "train", {"loss": train_loss, "acc": train_acc}, monitors["epoch"]
            )
            self.logger.info(f'-->Epoch : {monitors["epoch"]}')
            self.logger.info(
                f"  \u21b3train Loss/Acc = {train_loss:.4f}/{train_acc:.4f}"
            )

            # select best model
            if self.load_best_state and monitors[self.var_check] <= best_var:
                best_var = monitors[self.var_check]
                best_model_param = deepcopy(self.model.state_dict())

        writer.close()
        if not self.load_best_state:
            best_model_param = deepcopy(self.model.state_dict())

        # report the checkpoint time of end and compute cost time
        h, m, s = self.timer.stop()
        self.logger.info(f"[Train Finish] - [Cost Time = {h}H:{m}M:{s:.2f}S]")

        # load the best model and evaulate this model in testset
        self.model.load_state_dict(best_model_param)

        results = {}
        train_preds, train_target, train_loss = self.predict(train_loader)
        train_acc = self.get_acc(train_preds, train_target, train_ncls)
        results["train"] = {
            "preds": train_preds,
            "target": train_target,
            "acc": train_acc,
        }
        test_preds, test_target, test_loss = self.predict(test_loader)
        test_acc = self.get_acc(test_preds, test_target, train_ncls)
        results["test"] = {"preds": test_preds, "target": test_target, "acc": test_acc}

        self.logger.info(f"Loss: Train={train_loss:.4f} | Test={test_loss:.4f}")
        self.logger.info(f"Acc:  Train={train_acc:.4f} | Test={test_acc:.4f}")

        self.train_details["results"] = results
        self.train_details["best_model_param"] = best_model_param

        # store the training details
        train_details_path = os.path.join(log_dir, f"train_details.pt")
        torch.save(self.train_details, train_details_path)

        # store the best model parameters
        best_checkpoiont_path = os.path.join(log_dir, f"best_checkpoint.pth")
        torch.save(best_model_param, best_checkpoiont_path)

        return results


class ClassifierTwoStage(BaseClassifier):
    """Two-stage classifier training.

    Two-stage training strategy was used. In the first stage, the model was
    trained using only the training set with the early stopping criteria
    whereby the validation set accuracy and loss was monitored and training was
    stopped if there was no increase in the validation set accuracy (or loss)
    for consecutive 200 epochs. After reaching the stopping criteria, network
    parameters with the best validation set accuracy (or loss) were restored.
    In the second stage, the model was trained with the complete training data
    (train + validation set). The second stage training was stopped when the
    validation set loss reduced below the first stage training set loss.

    Parameters
    ----------
    model : Module
        Inherit Module and should define the forward method. The first
        parameter returned by model forward propagation is the prediction.
    loss_fn : str, Type[Module]
        Name of the loss function from `torch.nn` which will be used for
        training. If Module, means using a custom loss function.
        Note: custom loss_fn is a class (not an instance), and its
        initialization list is `(**loss_fn_args)`.
    loss_fn_args : dict, optional
        Additional arguments to be passed to the loss function.
    optimizer : str, Type[Optimizer]
        Name of the optimization function from torch.optim which will be used
        for training. If Optimizer, means using a custom optimizer.
        Note: custom optimizer is a class (not an instance), and its
        initialization list is `(model, lr=lr, **optimizer_args)`.
    optimizer_args : dict, optional
        Additional arguments to be passed to the optimization function.
    lr : float
        Learning rate.
    lr_sch : str, Type[LRScheduler], optional
        Name of the learning scheduler from `torch.optim.lr_scheduler` which
        will be used for training. If LRScheduler, means using a custom
        learning scheduler.
        Note: custom learning scheduler is a class (not an instance), and its
        initialization list is `(optimizer, **lr_sch_args)`.
    lr_sch_args : dict, optional
        Additional arguments to be passed to the lr_scheduler function.
    grad_acc : int
        Aradient accumulation.
    batch_size : int
        Mini-batch size.
    max_epochs_s1, max_epochs_s2 : int
        Maximum number of epochs in the x stage of training.
    no_increase_epochs : int
        Maximum number of consecutive epochs when the accuracy or loss of the
        first-stage validation set has no relative improvement.
    second_stage : bool
        If `True`, two-stage training will be performed.
    load_best_state : bool
        If `True`, two-stage will retrain based on the best state in first-
        stage. Otherwise, two-stage will retain based on the last state of the
        first-stage.
    var_check : str
        The best value (valid_inacc/valid_loss) to check while determining the
        best state which will be used for parameter initialization in the
        second stage of model training.
    cls_name : list of str
        The name of dataset labels.
    nGPU : int
        Select the gpu id to train. If the GPU is not available then the CPU is
        used.
    seed : int
        Select random seed for review.
    keep_data_gpu : bool
        Keep the dataset on the GPU to avoid the time consumption of data
        migration. Please adjust according to the personal GPU memory.
    data_size : tuple, list, optional
        Output the structure of the network model according to the input
        dimension if the `data_size` is given.
    depth : int, optional
        Depth of nested layers to display.
    verbose : int, str
        The log level of console. Default is INFO. Mainly used for debug.
    """

    def __init__(
        self,
        model: Module,
        loss_fn: str | Type[Module] = "NLLLoss",
        loss_fn_args: dict | None = None,
        optimizer: str | Type[Optimizer] = "Adam",
        optimizer_args: dict | None = None,
        lr: float = 0.001,
        lr_sch: str | Type[LRScheduler] | None = None,
        lr_sch_args: dict | None = None,
        grad_acc: int = 1,
        batch_size: int = 32,
        max_epochs_s1: int = 1500,
        max_epochs_s2: int = 600,
        no_increase_epochs: int = 200,
        second_stage: bool = True,
        var_check: Literal["valid_inacc", "valid_loss"] = "valid_inacc",
        load_best_state: bool = True,
        nGPU: int = 0,
        seed: int = DPEEG_SEED,
        keep_data_gpu: bool = True,
        depth: int | None = None,
        data_size: tuple | list | None = None,
        verbose: int | str = "INFO",
    ) -> None:
        super().__init__(
            model,
            loss_fn,
            loss_fn_args,
            optimizer,
            optimizer_args,
            lr,
            lr_sch,
            lr_sch_args,
            grad_acc,
            batch_size,
            nGPU,
            seed,
            keep_data_gpu,
            depth,
            data_size,
            verbose,
        )
        self.max_epochs_s1 = max_epochs_s1
        self.max_epochs_s2 = max_epochs_s2
        self.no_increase_epochs = no_increase_epochs
        self.second_stage = second_stage
        self.var_check = var_check
        self.load_best_state = load_best_state
        self.train_details.update(
            {
                "train_param": {
                    "max_epochs_s1": max_epochs_s1,
                    "max_epochs_s2": max_epochs_s2,
                    "no_increase_epochs": no_increase_epochs,
                    "second_stage": second_stage,
                    "var_check": var_check,
                    "load_best_state": load_best_state,
                }
            }
        )

    def fit(
        self,
        trainset: EEGData,
        validset: EEGData,
        testset: EEGData,
        log_dir: str,
    ):
        """Fit the model.

        Parameters
        ----------
        trainset : EEGData
            Dataset used for training.
        validset : EEGData
            Dataset used for validation.
        testset : EEGData
            Dataset used to evaluate the model.
        log_dir : str
            The path to save the training log.

        Returns
        -------
        dict
            Returns the training set, validation set, and test set results
            (including true labels, predicted labels and accuracy).
        """
        log_dir, writer = self._reset_fitter(log_dir)

        # check the best model
        best_var = float("inf")
        best_model_param = deepcopy(self.model.state_dict())
        best_optim_param = deepcopy(self.optimizer.state_dict())

        # initialize dataloader
        self.trainset = trainset
        self.validset = validset
        self.testset = testset
        train_loader = self.data_loader(trainset)
        valid_loader = self.data_loader(validset)
        test_loader = self.data_loader(testset)
        train_ncls = trainset.ncls
        valid_ncls = validset.ncls
        test_ncls = testset.ncls

        # start the training
        self.timer.start()
        self.logger.info(f"[Training...] - [{self.timer.ctime()}]")
        self.logger.info(
            f"[Train/Valid/Test] - "
            f"[{trainset.trials()}/{validset.trials()}/{testset.trials()}]"
        )

        stopcri = ComposeStopCriteria(
            {
                "Or": {
                    "cri1": {
                        "MaxEpoch": {
                            "max_epochs": self.max_epochs_s1,
                            "var_name": "epoch",
                        }
                    },
                    "cri2": {
                        "NoDecrease": {
                            "num_epochs": self.no_increase_epochs,
                            "var_name": self.var_check,
                        }
                    },
                }
            }
        )
        self.train_details["fit"] = {
            "type": "fit_with_val",
            "var_check": self.var_check,
            "stopcri_1": str(stopcri),
        }
        monitors = {
            "epoch": 0,
            "valid_loss": float("inf"),
            "valid_inacc": 1,
            "global_epoch": 0,
            "best_epoch": -1,
            "best_train_loss": float("inf"),
        }
        load_best_state = self.load_best_state
        early_stop_reached, do_stop = False, False

        while not do_stop:

            # train one epoch
            self.fit_epoch(train_loader)
            monitors["epoch"] += 1
            monitors["global_epoch"] += 1

            # evaluate the training and validation accuracy
            train_preds, train_target, train_loss = self.predict(train_loader)
            train_acc = self.get_acc(train_preds, train_target, train_ncls)
            valid_preds, valid_target, valid_loss = self.predict(valid_loader)
            valid_acc = self.get_acc(valid_preds, valid_target, valid_ncls)
            monitors["valid_inacc"] = 1 - valid_acc
            monitors["valid_loss"] = valid_loss

            # store loss and acc
            writer.add_scalars(
                "train",
                {"loss": train_loss, "acc": train_acc},
                monitors["global_epoch"],
            )
            writer.add_scalars(
                "valid",
                {"loss": valid_loss, "acc": valid_acc},
                monitors["global_epoch"],
            )
            # print the epoch info
            self.logger.info(f'-->Epoch : {monitors["epoch"]}')
            self.logger.info(
                f"  \u21b3train Loss/Acc = {train_loss:.4f}/{train_acc:.4f}"
                f" | valid Loss/Acc = {valid_loss:.4f}/{valid_acc:.4f}"
            )

            # select best model on Stage 1
            if load_best_state and monitors[self.var_check] <= best_var:
                best_var = monitors[self.var_check]
                best_model_param = deepcopy(self.model.state_dict())
                best_optim_param = deepcopy(self.optimizer.state_dict())
                monitors["best_train_loss"] = train_loss
                monitors["best_epoch"] = monitors["epoch"]

            # check whether to stop training
            if stopcri(monitors):
                # check whether to enter the second stage of training
                if self.second_stage and not early_stop_reached:
                    early_stop_reached = True
                    epoch = monitors["epoch"]

                    # load the best state
                    if load_best_state:
                        self.model.load_state_dict(best_model_param)
                        self.optimizer.load_state_dict(best_optim_param)
                        train_loss = monitors["best_train_loss"]
                        epoch = monitors["best_epoch"]

                    self.logger.info(
                        "[Early Stopping Reached] -> Training on full set."
                    )
                    self.logger.info(f"[Epoch = {epoch} | Loss = {train_loss:.4f}]")

                    # combine the train and valid dataset
                    train_loader = self.data_loader(trainset, validset)

                    # update stop monitor and epoch
                    stopcri = ComposeStopCriteria(
                        {
                            "Or": {
                                "cri1": {
                                    "MaxEpoch": {
                                        "max_epochs": self.max_epochs_s2,
                                        "var_name": "epoch",
                                    }
                                },
                                "cri2": {
                                    "Smaller": {
                                        "var": train_loss,
                                        "var_name": "valid_loss",
                                    }
                                },
                            }
                        }
                    )
                    self.train_details["fit"]["stopcri_2"] = str(stopcri)
                    monitors["epoch"] = 0
                    load_best_state = False
                elif self.second_stage and early_stop_reached:
                    do_stop = True
                    best_model_param = deepcopy(self.model.state_dict())
                # no second stage
                else:
                    do_stop = True
                    if not load_best_state:
                        best_model_param = deepcopy(self.model.state_dict())

        writer.close()

        # report the checkpoint time of end and compute cost time
        h, m, s = self.timer.stop()
        self.logger.info(f"[Train Finish] - [Cost Time = {h}H:{m}M:{s:.2f}S]")

        # load the best model and evaulate this model in testset
        self.model.load_state_dict(best_model_param)

        results = {}
        train_preds, train_target, train_loss = self.predict(train_loader)
        train_acc = self.get_acc(train_preds, train_target, train_ncls)
        results["train"] = {
            "preds": train_preds,
            "target": train_target,
            "acc": train_acc,
        }
        valid_preds, valid_target, valid_loss = self.predict(valid_loader)
        valid_acc = self.get_acc(valid_preds, valid_target, valid_ncls)
        results["valid"] = {
            "preds": valid_preds,
            "target": valid_target,
            "acc": valid_acc,
        }
        test_preds, test_target, test_loss = self.predict(test_loader)
        test_acc = self.get_acc(test_preds, test_target, test_ncls)
        results["test"] = {
            "preds": test_preds,
            "target": test_target,
            "acc": test_acc,
        }

        self.logger.info(
            f"Loss: Train={train_loss:.4f} | Valid={valid_loss:.4f} | "
            f"test={test_loss:.4f}"
        )
        self.logger.info(
            f"Acc:  Train={train_acc:.4f} | Valid={valid_acc:.4f} | "
            f"Test={test_acc:.4f}"
        )

        self.train_details["results"] = results
        self.train_details["best_model_param"] = best_model_param

        # store the training details
        train_details_path = os.path.join(log_dir, f"train_details.pt")
        torch.save(self.train_details, train_details_path)

        # store the best model
        best_checkpoiont_path = os.path.join(log_dir, f"best_checkpoint.pth")
        torch.save(best_model_param, best_checkpoiont_path)

        return results



import torch
import torch.nn as nn

from .utils import Conv2dWithNorm, LinearWithNorm
from ..tools.docs import fill_doc


__all__ = ["ShallowConvNet"]


class Lambda(nn.Module):
    def __init__(self, func) -> None:
        super().__init__()
        self.func = func

    def forward(self, x):
        return self.func(x)


@fill_doc
class ShallowConvNet(nn.Module):
    """Deep Learning With Convolutional Neural Networks for EEG Decoding and
    Visualization (ShallowConvNet).

    Shallow ConvNet [1]_, inspired by the FBCSP pipeline, is specifically
    tailored to decode band power features. The transformations performed by
    the shallow ConvNet are similar to the transformations of FBCSP.
    Concretely, the first two layers of the shallow ConvNet perform a temporal
    convolution and a spatial filter, as in the deep ConvNet. These steps are
    analogous to the bandpass and CSP spatial filter steps in FBCSP. In
    contrast to the deep ConvNet, the temporal convolution of the shallow
    ConvNet had a larger kernel size, allowing a larger range of
    transformations in this layer (smaller kernel sizes for the shallow ConvNet
    led to lower accuracies in preliminary experiments on the training set).
    After the temporal convolution and the spatial filter of the shallow
    ConvNet, a squaring nonlinearity, a mean pooling layer and a logarithmic
    activation function followed; together these steps are analogous to the
    trial log-variance computation in FBCSP. In contrast to FBCSP, the shallow
    ConvNet embeds all the computational steps in a single network, and thus
    all steps can be optimized jointly. Also, due to having several pooling
    regions within one trial, the shallow ConvNet can learn a temporal
    structure of the band power changes within the trial.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    F : int
        The number of convolution channels.
    C : int
        Temporal convolution kernel size.
    P : int
        Pooling kernel size.
    S : int
        Pooling layer stride size.
    dropout : float
        Dropout rate.

    References
    ----------
    .. [1] R. T. Schirrmeister et al., “Deep learning with convolutional neural
        networks for EEG decoding and visualization,” Human Brain Mapping,
        vol. 38, no. 11, pp.5391-5420, 2017, doi: 10.1002/hbm.23730.
    """

    def __init__(
        self,
        nCh: int,
        nTime: int,
        nCls: int,
        F: int = 40,
        C: int = 14,
        P: int = 35,
        S: int = 7,
        dropout: float = 0.5,
    ) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime

        self.conv = nn.Sequential(
            Conv2dWithNorm(1, F, (1, C), max_norm=2, bias=False),
            Conv2dWithNorm(F, F, (nCh, 1), max_norm=2, bias=False, groups=F),
            nn.BatchNorm2d(F),
            Lambda(torch.square),
            nn.AvgPool2d((1, P), stride=(1, S)),
            Lambda(torch.log),
        )

        linear_in = self.forward_flatten().shape[1]
        self.head = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(dropout),
            LinearWithNorm(linear_in, nCls, max_norm=0.5),
            nn.LogSoftmax(dim=1),
        )

    def forward_flatten(self):
        x = torch.rand(1, 1, self.nCh, self.nTime)
        x = self.conv(x)
        x = torch.flatten(x, 1, -1)
        return x

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, bands, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, nCls)`.
        """
        x = self.conv(x)
        x = self.head(x)
        return x



import torch
import torch.nn as nn

from .utils import Conv2dWithNorm, LinearWithNorm
from ..tools.docs import fill_doc


__all__ = ["DeepConvNet"]


@fill_doc
class DeepConvNet(nn.Module):
    """Deep Learning With Convolutional Neural Networks for EEG Decoding and
    Visualization (Deep ConvNet).

    Deep ConvNet [1]_ had four convolution-max-pooling blocks, with a special
    first block designed to handle EEG input, followed by three standard
    convolutionmax-pooling blocks and a dense softmax classification layer.
    The first convolutional block was split into two layers in order to better
    handle the large number of input channelsone input channel per electrode
    compared to three input channels (one per color) in rgb-images. In the
    first layer, each filter performs a convolution over time, and in the
    second layer, each filter performs a spatial filtering with weights for all
    possible pairs of electrodes with filters of the preceding temporal
    convolution. Note that as there is no activation function in between the
    two layers, they could in principle be combined into one layer. Using two
    layers however implicitly regularizes the overall convolution by forcing a
    separation of the linear transformation into a combination of a temporal
    convolution and a spatial filter.

    Parameters
    ----------
    %(nCh)s
    %(nTime)s
    %(nCls)s
    dropout : float
        Dropout rate.

    References
    ----------
    .. [1] R. T. Schirrmeister et al., “Deep learning with convolutional neural
        networks for EEG decoding and visualization,” Human Brain Mapping,
        vol. 38, no. 11, pp.5391-5420, 2017, doi: 10.1002/hbm.23730.
    """

    def __init__(self, nCh: int, nTime: int, nCls: int, dropout: float = 0.25) -> None:
        super().__init__()
        self.nCh = nCh
        self.nTime = nTime
        kernel_size = [1, 10]
        filter_layer = [25, 50, 100, 200]

        first_layer = nn.Sequential(
            Conv2dWithNorm(1, 25, kernel_size, max_norm=2),
            Conv2dWithNorm(25, 25, (nCh, 1), bias=False, max_norm=2),
            nn.BatchNorm2d(25),
            nn.ELU(),
            nn.MaxPool2d((1, 3)),
        )
        middle_layer = nn.Sequential(
            *[
                nn.Sequential(
                    nn.Dropout(dropout),
                    Conv2dWithNorm(in_f, out_f, kernel_size),
                    nn.BatchNorm2d(out_f),
                    nn.ELU(),
                    nn.MaxPool2d((1, 3)),
                )
                for in_f, out_f in zip(filter_layer, filter_layer[1:])
            ]
        )
        self.conv_layer = nn.Sequential(first_layer, middle_layer)

        linear_in = self._forward_flatten().shape[1]
        self.head = nn.Sequential(
            nn.Flatten(),
            LinearWithNorm(linear_in, nCls, max_norm=0.5),
            nn.LogSoftmax(dim=1),
        )

    def _forward_flatten(self):
        x = torch.rand(1, 1, self.nCh, self.nTime)
        x = self.conv_layer(x)
        x = torch.flatten(x, start_dim=1, end_dim=-1)
        return x

    def forward(self, x):
        """Forward pass function that processes the input EEG data and produces
        the decoded results.

        Parameters
        ----------
        x : Tensor
            Input EEG data, shape `(batch_size, 1, nCh, nTime)`.

        Returns
        -------
        cls_prob : Tensor
            Predicted class probability, shape `(batch_size, cls)`.
        """
        x = self.conv_layer(x)
        x = self.head(x)
        return x



import numpy as np
from scipy.io import loadmat
from mne.channels import make_standard_montage
from mne.io import RawArray
from mne import create_info

from .base import EpochsDataset, DATA_PATH
from .download import data_dl
from ..utils import get_init_args
from ..tools.docs import fill_doc


URL = "ftp://parrot.genomics.cn/gigadb/pub/"


@fill_doc
class OpenBMI_MI(EpochsDataset):
    """BMI/OpenBMI dataset for MI.

    .. admonition:: Dataset summary

        ====== ====== ======= ===== ======== ======
        Subj   Chan   Time    Cls   Freq     Sess
        ====== ====== ======= ===== ======== ======
        54     62     4.0 s   2     1000 Hz  2
        ====== ====== ======= ===== ======== ======

    This data set consists of EEG data from 54 subjects of a study published in
    [1]_. EEG signals were recorded with a sampling rate of 1,000 Hz and
    collected with 62 Ag/AgCl electrodes. The EEG amplifier used in the
    experiment was a BrainAmp (Brain Products; Munich, Germany). The channels
    were nasion-referenced and grounded to electrode AFz. Additionally, an EMG
    electrode recorded from each flexor digitorum profundus muscle with the
    olecranon used as reference. The impedances of the EEG electrodes were
    maintained below 10 k during the entire experiment.

    MI paradigm The MI paradigm was designed following a well-established
    system protocol. For all blocks, the first 3 s of each trial began with a
    black fixation cross that appeared at the center of the monitor to prepare
    subjects for the MI task. Afterwards, the subject performed the imagery
    task of grasping with the appropriate hand for 4 s when the right or left
    arrow appeared as a visual cue. After each task, the screen remained blank
    for 6 s (± 1.5 s). The experiment consisted of training and test phases;
    each phase had 100 trials with balanced right and left hand imagery tasks.
    During the online test phase, the fixation cross appeared at the center of
    the monitor and moved right or left, according to the real-time classifier
    output of the EEG signal.

    References
    ----------
    .. [1] Lee, M. H., Kwon, O. Y., Kim, Y. J., Kim, H. K., Lee, Y. E.,
        Williamson, J., … Lee, S. W. (2019). EEG dataset and OpenBMI toolbox
        for three BCI paradigms: An investigation into BCI illiteracy.
        GigaScience, 8(5), 1-16. https://doi.org/10.1093/gigascience/giz002

    Parameters
    ----------
    %(subjects)s
    %(epochs_tmin_tmax)s
    %(baseline)s
    %(picks)s
    %(resample)s
    %(rename)s
    """

    def __init__(
        self,
        subjects: list[int] | None = None,
        tmin: float = 0,
        tmax: float | None = None,
        baseline: tuple[int, int] | None = None,
        picks: list[str] | None = None,
        resample: float | None = None,
        rename: str | None = None,
    ) -> None:
        super().__init__(
            repr=get_init_args(self, locals(), rename=rename, ret_dict=True),
            subject_list=list(range(1, 55)),
            interval=[0.0, 4.0],
            event_id={"right_hand": 1, "left_hand": 2},
            subjects=subjects,
            tmin=tmin,
            tmax=tmax,
            baseline=baseline,
            picks=picks,
            resample=resample,
        )
        self._data_url = f"{URL}10.5524/100001_101000/100542/"
        self._data_path = DATA_PATH / "openbmi_mi"

    def _load_run(self, data, verbose):
        sfreq = data["fs"].item()

        eeg_ch_names = [np.squeeze(c).item() for c in np.ravel(data["chan"])]
        eeg_ch_types = ["eeg"] * len(eeg_ch_names)
        emg_ch_names = [np.squeeze(c).item() for c in np.ravel(data["EMG_index"])]
        emg_ch_types = ["emg"] * len(emg_ch_names)
        raw_ch_names = eeg_ch_names + emg_ch_names + ["STI 014"]
        raw_ch_types = eeg_ch_types + emg_ch_types + ["stim"]
        info = create_info(raw_ch_names, sfreq, raw_ch_types)  # type: ignore

        eeg = data["x"].transpose(1, 0) * 1e-6  # to load the signal in Volts
        emg = data["EMG"].transpose(1, 0) * 1e-6  # to load the signal in Volts
        event_times_in_samples = data["t"].squeeze()
        event_id = data["y_dec"].squeeze()
        stim = np.zeros(len(data["x"]))
        for i_sample, id_class in zip(event_times_in_samples, event_id):
            stim[i_sample] += id_class
        raw_data = np.concatenate([eeg, emg, stim[None, :]])

        raw = RawArray(data=raw_data, info=info, verbose=verbose)
        raw.set_montage(make_standard_montage("standard_1005"))

        return raw

    def _load_sess(self, filename, verbose):
        mat = loadmat(filename)
        train = self._load_run(mat["EEG_MI_train"][0, 0], verbose)
        test = self._load_run(mat["EEG_MI_test"][0, 0], verbose)
        return {"train": train, "test": test}

    def _get_single_subject_raw(self, subject: int, verbose="ERROR"):
        sessions = {}
        for r in [1, 2]:
            filename = data_dl(
                f"{self._data_url}session{r}/s{subject}/sess{r:02d}_subj{subject:02d}_EEG_MI.mat",
                self._data_path,
                force_update=False,
            )
            sessions[f"session_{r}"] = self._load_sess(filename, verbose)

        return sessions



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from pathlib import Path

import mne
import numpy as np
from scipy.io import loadmat
from numpy import ndarray
from mne.channels import make_standard_montage
from mne.io import RawArray
from mne import create_info

from .download import data_dl
from ..utils import DPEEG_DIR


GIGADB_URL = "ftp://parrot.genomics.cn/gigadb/pub/"
PHYSIONET_URL = "https://physionet.org/files/"
BNCI_URL = "http://bnci-horizon-2020.eu/database/data-sets/"
GIN_URL = "https://web.gin.g-node.org/"


def load_data(
    subject,
    dataset: str,
    path: str | None = None,
    force_update: bool = False,
):
    """Return the data of a single subject.

    The returned data is a dictionary with the folowing structure:

        data = {'session_id':
                    {'run_id': raw}
                }

    Parameters
    ----------
    subject : int
        The subject to load.
    dataset : string
        The dataset name.
    path : str, optional
        Location of where to look for the data storing location. Default is
        `~/dpeeg/dataset`. If the dataset is not found under the given path,
        the data will be automatically downloaded to the specified folder.
    force_update : bool
        Force update of the dataset even if a local copy exists.

    Returns
    -------
    raws : dict
        dict containing the raw data.
    """
    dataset_list = {
        "bciciv2a": {"url": BNCI_URL, "load": _load_bciciv2a},
        "bciciv2b": {"url": BNCI_URL, "load": _load_bciciv2b},
        "openbmi_mi": {
            "url": f"{GIGADB_URL}10.5524/100001_101000/100542/",
            "load": _load_openbmi_mi,
        },
        "high_gamma": {
            "url": f"{GIN_URL}robintibor/high-gamma-dataset/raw/master/data",
            "load": _load_high_gamma,
        },
    }

    if dataset not in dataset_list.keys():
        raise ValueError(
            f"Dataset {dataset} is not available. "
            f"Valid dataset are {', '.join(dataset_list.keys())}."
        )

    dl_path = Path(DPEEG_DIR, "datasets") if path is None else Path(path)
    return dataset_list[dataset]["load"](
        subject, dl_path / dataset, dataset_list[dataset]["url"], force_update
    )


def _check_subject(subject: int, total_subject: int):
    if (subject < 1) or (subject > total_subject):
        raise ValueError(
            f"Subject must be between 1 and {total_subject}. Got {subject}."
        )


def standardize_keys(d):
    master_list = [
        ["both feet", "feet"],
        ["left hand", "left_hand"],
        ["right hand", "right_hand"],
        ["FEET", "feet"],
        ["HAND", "right_hand"],
        ["NAV", "navigation"],
        ["SUB", "subtraction"],
        ["WORD", "word_ass"],
    ]
    for old, new in master_list:
        if old in d.keys():
            d[new] = d.pop(old)


def _convert_run(run, ch_names, ch_types):
    """Convert one run to raw."""
    event_id = {}
    n_chan = run.X.shape[1]
    montage = make_standard_montage("standard_1005")
    eeg_data = 1e-6 * run.X
    sfreq = run.fs

    if not ch_names:
        ch_names = [f"EEG{ch}" for ch in range(1, n_chan + 1)]
        montage = None

    if not ch_types:
        ch_types = ["eeg"] * n_chan

    trigger = np.zeros((len(eeg_data), 1))
    # some runs does not contains trials i.e baseline runs
    if len(run.trial) > 0:
        trigger[run.trial - 1, 0] = run.y
    else:
        return None, None

    eeg_data = np.c_[eeg_data, trigger]
    ch_names = ch_names + ["stim"]
    ch_types = ch_types + ["stim"]
    event_id = {ev: (ii + 1) for ii, ev in enumerate(run.classes)}
    info = create_info(ch_names=ch_names, ch_types=ch_types, sfreq=sfreq)
    raw = RawArray(data=eeg_data.T, info=info, verbose=False)
    raw.set_montage(montage)
    return raw, event_id


def _convert_mi(filename, ch_names, ch_types):
    runs = []
    event_id = {}
    data = loadmat(filename, struct_as_record=False, squeeze_me=True)

    if isinstance(data["data"], ndarray):
        run_array = data["data"]
    else:
        run_array = [data["data"]]

    for run in run_array:
        raw, evd = _convert_run(run, ch_names, ch_types)
        if raw is None:
            continue
        runs.append(raw)
        event_id.update(evd)  # type: ignore
        # change labels to match rest
    standardize_keys(event_id)
    return runs, event_id


def _load_bciciv2a(
    subject: int,
    path: str,
    base_url: str,
    force_update: bool = False,
):
    """Load data set 2a of the BCI Competition IV."""
    _check_subject(subject, 9)

    # fmt: off
    ch_names = [
        "Fz", "FC3", "FC1", "FCz", "FC2", "FC4", "C5", "C3", "C1", "Cz", "C2",
        "C4", "C6", "CP3", "CP1", "CPz", "CP2", "CP4", "P1", "Pz", "P2", "POz",
        "EOG1", "EOG2", "EOG3",
    ]
    # fmt: on
    ch_types = ["eeg"] * 22 + ["eog"] * 3

    sessions = {}
    for i, r in enumerate(["T", "E"], start=1):
        url = f"{base_url}001-2014/A{subject:02d}{r}.mat"
        filename = data_dl(url, path, force_update)
        runs, _ = _convert_mi(filename, ch_names, ch_types)
        sessions[f"session_{i}"] = {
            f"run_{ii}": run for ii, run in enumerate(runs, start=1)
        }
    return sessions


def _load_bciciv2b(
    subject: int,
    path: str,
    base_url: str,
    force_updata: bool = False,
):
    """Load data set 2b of the BCI Competition IV."""
    _check_subject(subject, 9)

    ch_names = ["C3", "Cz", "C4", "EOG1", "EOG2", "EOG3"]
    ch_types = ["eeg"] * 3 + ["eog"] * 3

    sessions = []
    for i, r in enumerate(["T", "E"], start=1):
        url = f"{base_url}004-2014/B{subject:02d}{r}.mat"
        filename = data_dl(url, path, force_updata)
        runs, _ = _convert_mi(filename, ch_names, ch_types)
        sessions.extend(runs)

    sessions = {
        f"session_{ii}": {"run_1": run} for ii, run in enumerate(sessions, start=1)
    }
    return sessions


def _load_openbmi_mi_run(data):
    sfreq = data["fs"].item()

    eeg_ch_names = [np.squeeze(c).item() for c in np.ravel(data["chan"])]
    eeg_ch_types = ["eeg"] * len(eeg_ch_names)
    emg_ch_names = [np.squeeze(c).item() for c in np.ravel(data["EMG_index"])]
    emg_ch_types = ["emg"] * len(emg_ch_names)
    raw_ch_names = eeg_ch_names + emg_ch_names + ["STI 014"]
    raw_ch_types = eeg_ch_types + emg_ch_types + ["stim"]
    info = create_info(raw_ch_names, sfreq, raw_ch_types)  # type: ignore

    eeg = data["x"].transpose(1, 0) * 1e-6  # to load the signal in Volts
    emg = data["EMG"].transpose(1, 0) * 1e-6  # to load the signal in Volts
    event_times_in_samples = data["t"].squeeze()
    event_id = data["y_dec"].squeeze()
    stim = np.zeros(len(data["x"]))
    for i_sample, id_class in zip(event_times_in_samples, event_id):
        stim[i_sample] += id_class
    raw_data = np.concatenate([eeg, emg, stim[None, :]])

    raw = RawArray(data=raw_data, info=info, verbose=False)
    raw.set_montage(make_standard_montage("standard_1005"))

    return raw


def _load_openbmi_mi_sess(filename):
    mat = loadmat(filename)
    train = _load_openbmi_mi_run(mat["EEG_MI_train"][0, 0])
    test = _load_openbmi_mi_run(mat["EEG_MI_test"][0, 0])
    return {"train": train, "test": test}


def _load_openbmi_mi(
    subject: int,
    path: str,
    base_url: str,
    force_update: bool = False,
):
    """Load data set MI of the OpenBMI."""
    _check_subject(subject, 54)

    sessions = {}
    for r in [1, 2]:
        url = (
            f"{base_url}session{r}/s{subject}/sess{r:02d}_"
            f"subj{subject:02d}_EEG_MI.mat"
        )
        filename = data_dl(url, path, force_update)
        sessions[f"session_{r}"] = _load_openbmi_mi_sess(filename)

    return sessions


def _load_high_gamma(
    subject: int,
    path: str,
    base_url: str,
    force_update: bool = False,
):
    """Load data set of the High-Gamma Dataset."""
    _check_subject(subject, 14)

    sessions = {}
    montage = make_standard_montage("standard_1005")
    for i, r in enumerate(["train", "test"], start=1):
        filename = data_dl(f"{base_url}/{r}/{subject:d}.edf", path, force_update)
        raw = mne.io.read_raw_edf(
            filename, infer_types=True, preload=True, verbose=False
        )
        raw.set_montage(montage)
        sessions[f"session_{i}"] = raw

    return sessions



# Authors: SheepTAO <sheeptao@outlook.com>

# License: MIT
# Copyright the dpeeg contributors.

from abc import ABC, abstractmethod
from pathlib import Path

from ..tools.logger import Logger
from ..tools.timer import Timer


class Trainer(ABC):
    """Top-level trainer.

    Rewrite the trainer ``fit`` function according to different training
    methods.

    Attributes
    ----------
    logger : Logger
        The log manager of the trainer. The log manager is updated by calling
        ``_reset_logger`` function when the log file output location needs to
        be updated.
    timer : Timer
        The default timer for the trainer. Other timers can be registered and
        used as needed.
    """

    def __init__(self, model, verbose) -> None:
        super().__init__()
        self.model = model
        self.verbose = verbose
        self.logger = Logger("dpeeg_trainer", clevel=verbose)
        self.timer = Timer()

    def _reset_logger(self, log_dir: str | Path):
        log_dir = Path(log_dir).absolute()
        self.logger = Logger(
            str(log_dir),
            path=log_dir / "running.log",
            flevel=self.verbose,
            clevel=self.verbose,
        )

    @abstractmethod
    def fit(*args, **kwargs) -> dict:
        """The top-level fitting function can be overridden by different
        trainers as needed and returns the fitting results in a dictionary.
        """
        pass



# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

import sys

from mne.utils._bunch import BunchConst


##############################################################################
# Define our standard documentation entries
#
# To reduce redundancy across functions, please standardize the format to
# ``argument_optional_keywords``. For example ``tmin_raw`` for an entry that
# is specific to ``raw`` and since ``tmin`` is used other places, needs to
# be disambiguated. This way the entries will be easy to find since they
# are alphabetized (you can look up by the name of the argument). This way
# the same ``docdict`` entries are easier to reuse.

docdict = BunchConst()

# fmt: off

# %%
# A

docdict["aug_only_train"] = """
only_train : bool
    If True, data augmentation is performed only on the training set.
"""

docdict["aug_strict"] = """
strict : bool
    If False, allow the input data to be unsplit. In this case, data
    augmentation will be applied to all data. Please ensure that this operation
    does not cause data leakage according to the method of data augmentation.
"""

# %%
# B

docdict["baseline"] = """
baseline : tuple of int, optional
    The time interval to consider as “baseline” when applying baseline
    correction. If ``None``, do not apply baseline correction.
"""

# %%
# E

docdict["eegdata_set"] = """
eegdata : :ref:`eeg_data`, :ref:`eeg_dataset`
    The basic types are eegdata and eegdataset.
"""

docdict["trans_ret"] = """

"""

# %%
# N

docdict["nCh"] = """
nCh : int
    Number of electrode channels.
"""

docdict["nTime"] = """
nTime : int
    Number of data sampling points. For example, a 4-second data input with a
    sampling rate of 250 Hz is 1000.
"""

docdict["nCls"] = """
nCls : int
    Number of classification categories.
"""

# %%
# O

docdict["out_folder"] = """
out_folder : str, optional
    Store all experimental results in a folder named with the model class name 
    in the specified folder ('~/dpeeg/out/model/exp/dataset/timestamp').
"""

# %%
# P

docdict["picks"] = """
picks : list of str, optional
    Channels to include. If ``None``, pick all channels.
"""

# %%
# R

docdict["resample"] = """
resample : float, optional
    Resample data.
"""

docdict["rename"] = """
rename : str, optional
    Rename the dataset.
"""

# %%
# S

docdict["subjects"] = """
subjects : list of int, optional
    List of subject number. If ``None``, all subjects will be loaded.
"""

# %%
# T

docdict["epochs_tmin_tmax"] = """
tmin-tmax : float
    Start and end time of the epochs in seconds, relative to the time locked 
    event. The closest or matching samples corresponding to the start and end 
    time are included. Default is start and end time of epochs, respectively.
"""

docdict["trainer"] = """
trainer : Trainer
    Trainer used for training module on dataset.
"""

docdict["timestamp"] = """
timestamp : bool
    Output folders are timestamped.
"""

docdict["train_for_test"] = """
train_for_test : bool
    If True, normalize the test set using the values in the training set.
"""

docdict["raw_tmin_tmax"] = """
tmin-tmax : float
    Start and end time of the raw in seconds. Default to use all data.
"""

# fmt: on

docdict_indented = {}


def fill_doc(f):
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of. Will be modified in place.

    Returns
    -------
    f : callable
        The function, potentially with an updated ``__doc__``.
    """
    docstring = f.__doc__
    if not docstring:
        return f
    lines = docstring.splitlines()
    # Find the minimum indent of the main docstring, after first line
    if len(lines) < 2:
        icount = 0
    else:
        icount = _indentcount_lines(lines[1:])
    # Insert this indent to dictionary docstrings
    try:
        indented = docdict_indented[icount]
    except KeyError:
        indent = " " * icount
        docdict_indented[icount] = indented = {}
        for name, dstr in docdict.items():
            lines = dstr.splitlines()
            try:
                newlines = [lines[0]]
                for line in lines[1:]:
                    newlines.append(indent + line)
                indented[name] = "\n".join(newlines)
            except IndexError:
                indented[name] = dstr
    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split("\n")[0] if funcname is None else funcname
        raise RuntimeError(f"Error documenting {funcname}:\n{exp}")
    return f


def _indentcount_lines(lines):
    """Compute minimum indent for all lines in line list."""
    indentno = sys.maxsize
    for line in lines:
        stripped = line.lstrip()
        if stripped:
            indentno = min(indentno, len(line) - len(stripped))
    if indentno == sys.maxsize:
        return 0
    return indentno
